"""
SWAN Components Group Module

This module provides group components for organizing SWAN model configurations in the ROMPY framework.
"""

import logging
from typing import Annotated, Literal, Optional, Union

from pydantic import Field, field_validator, model_validator

from rompy_swan.components.base import BaseComponent
from rompy_swan.components.inpgrid import CURVILINEAR, ICE, REGULAR, UNSTRUCTURED, WIND
from rompy_swan.components.lockup import COMPUTE_NONSTAT, COMPUTE_STAT, STOP
from rompy_swan.components.output import (
    BLOCK,
    BLOCKS,
    CURVES,
    FRAME,
    GROUP,
    ISOLINE,
    NESTOUT,
    NGRID,
    NGRID_UNSTRUCTURED,
    OUTPUT_OPTIONS,
    POINTS,
    POINTS_FILE,
    QUANTITIES,
    RAY,
    SPECIAL_NAMES,
    SPECOUT,
    TABLE,
    TEST,
)
from rompy_swan.components.physics import (
    BRAGG,
    BRAGG_FILE,
    BRAGG_FT,
    BREAKING_BKD,
    BREAKING_CONSTANT,
    DIFFRACTION,
    FRICTION_COLLINS,
    FRICTION_JONSWAP,
    FRICTION_MADSEN,
    FRICTION_RIPPLES,
    GEN1,
    GEN2,
    GEN3,
    LIMITER,
    MUD,
    NEGATINP,
    OBSTACLES,
    OFFS,
    QUADRUPL,
    SCAT,
    SETUP,
    SICE,
    SICE_D15,
    SICE_M18,
    SICE_R19,
    SICE_R21B,
    SSWELL_ARDHUIN,
    SSWELL_ROGERS,
    SSWELL_ZIEGER,
    SURFBEAT,
    TRIAD,
    TRIAD_DCTA,
    TRIAD_LTA,
    TRIAD_SPB,
    TURBULENCE,
    VEGETATION,
    WCAPPING_AB,
    WCAPPING_KOMEN,
)
from rompy_swan.components.startup import COORDINATES, MODE, PROJECT, SET
from rompy_swan.types import PhysicsOff

logger = logging.getLogger(__name__)


class BaseGroupComponent(BaseComponent):
    """Base group component.

    Base class for SWAN group components. This base class modifies the `render()`
    method to allow rendering of a list of components. It is not intended to be used
    directly but to be subclassed by other group components.

    """

    model_type: Literal["group", "GROUP"] = Field(
        default="group", description="Model type discriminator"
    )

    def render(self, *args, **kwargs) -> str:
        """Override base class to allow rendering list of components."""
        cmds = []
        for cmd in self.cmd():
            cmds.append(super().render(cmd))
        return "\n\n".join(cmds)


# =====================================================================================
# Startup
# =====================================================================================
PROJECT_TYPE = Annotated[PROJECT, Field(description="Project component")]
SET_TYPE = Annotated[SET, Field(description="Set component")]
MODE_TYPE = Annotated[MODE, Field(description="Mode component")]
COORDINATES_TYPE = Annotated[COORDINATES, Field(description="Coordinates component")]


class STARTUP(BaseGroupComponent):
    """Startup group component.

    .. code-block:: text

        PROJECT ...
        SET ...
        MODE ...
        COORDINATES ...

    This group component is used to group individual startup components. Only fields
    that are explicitly prescribed are rendered by this group component.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.startup import PROJECT, SET, MODE, COORDINATES
        from rompy_swan.components.group import STARTUP
        proj = PROJECT(nr="01")
        set = SET(level=0.5, direction_convention="nautical")
        mode = MODE(kind="nonstationary", dim="twodimensional")
        coords = COORDINATES(kind=dict(model_type="spherical", projection="ccm"))
        startup = STARTUP(
            project=proj,
            set=set,
            mode=mode,
            coordinates=coords,
        )
        print(startup.render())

    """

    model_type: Literal["startup", "STARTUP"] = Field(
        default="startup", description="Model type discriminator"
    )
    project: Optional[PROJECT_TYPE] = Field(default=None)
    set: Optional[SET_TYPE] = Field(default=None)
    mode: Optional[MODE_TYPE] = Field(default=None)
    coordinates: Optional[COORDINATES_TYPE] = Field(default=None)

    def cmd(self) -> str:
        """Command file string for this component."""
        repr = []
        if self.project is not None:
            repr += [f"{self.project.cmd()}"]
        if self.set is not None:
            repr += [f"{self.set.cmd()}"]
        if self.mode is not None:
            repr += [f"{self.mode.cmd()}"]
        if self.coordinates is not None:
            repr += [f"{self.coordinates.cmd()}"]
        return repr


# =====================================================================================
# Inpgrid
# =====================================================================================
INPGRID_TYPE = Annotated[
    Union[REGULAR, CURVILINEAR, UNSTRUCTURED, WIND, ICE],
    Field(discriminator="model_type"),
]


class INPGRIDS(BaseGroupComponent):
    """SWAN input grids group component.

    .. code-block:: text

        INPGRID ...
        READGRID ...

        INPGRID ...
        READGRID ...

        ...

    This group component is a convenience to allow defining and rendering
    a list of input grid components.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.inpgrid import REGULAR, ICE
        from rompy_swan.components.group import INPGRIDS
        inpgrid_bottom = REGULAR(
            grid_type="bottom",
            excval=-99.0,
            xpinp=172.0,
            ypinp=-41.0,
            alpinp=0.0,
            mxinp=99,
            myinp=99,
            dxinp=0.005,
            dyinp=0.005,
            readinp=dict(fname1="bottom.txt"),
        )
        inpgrid_wind = REGULAR(
            grid_type="wind",
            excval=-99.0,
            xpinp=172.0,
            ypinp=-41.0,
            alpinp=0.0,
            mxinp=99,
            myinp=99,
            dxinp=0.005,
            dyinp=0.005,
            readinp=dict(fname1="wind.txt"),
            nonstationary=dict(
                tbeg="2019-01-01T00:00:00",
                tend="2019-01-07 00:00:00",
                delt=3600,
                dfmt="hr",
            ),
        )
        inpgrid_ice_cte = ICE(aice=0.8, hice=2.0)
        inpgrids = INPGRIDS(inpgrids=[inpgrid_bottom, inpgrid_wind, inpgrid_ice_cte])
        print(inpgrids.render())

    """

    model_type: Literal["inpgrids"] = Field(
        default="inpgrids", description="Model type discriminator"
    )
    inpgrids: list[INPGRID_TYPE] = Field(
        min_length=1,
        description="List of input grid components",
    )

    @field_validator("inpgrids")
    @classmethod
    def ensure_unique_grid_type(cls, inpgrids: INPGRID_TYPE) -> INPGRID_TYPE:
        """Ensure that each grid type is unique."""
        grid_types = [inp.grid_type for inp in inpgrids if hasattr(inp, "grid_type")]
        if len(grid_types) != len(set(grid_types)):
            raise ValueError("Each grid type must be unique")
        return inpgrids

    def cmd(self) -> str | list:
        repr = []
        for inpgrid in self.inpgrids:
            repr += [inpgrid.cmd()]
        return repr


# =====================================================================================
# Physics
# =====================================================================================
NEGATINP_TYPE = Annotated[NEGATINP, Field(description="Negative wind input component")]
QUADRUPL_TYPE = Annotated[QUADRUPL, Field(description="Quadruplets component")]
VEGETATION_TYPE = Annotated[VEGETATION, Field(description="Vegetation component")]
MUD_TYPE = Annotated[MUD, Field(description="Mud component")]
TURBULENCE_TYPE = Annotated[TURBULENCE, Field(description="Turbulence component")]
LIMITER_TYPE = Annotated[LIMITER, Field(description="Limiter component")]
OBSTACLE_TYPE = Annotated[OBSTACLES, Field(description="Obstacle group component")]
SETUP_TYPE = Annotated[SETUP, Field(description="Setup component")]
DIFFRACTION_TYPE = Annotated[DIFFRACTION, Field(description="Diffraction component")]
SURFBEAT_TYPE = Annotated[SURFBEAT, Field(description="Surfbeat component")]
SCAT_TYPE = Annotated[SCAT, Field(description="Scattering component")]
OFF_TYPE = Annotated[OFFS, Field(description="Deactivate components")]
GEN_TYPE = Annotated[
    Union[GEN1, GEN2, GEN3],
    Field(description="Wave generation component", discriminator="model_type"),
]
SSWELL_TYPE = Annotated[
    Union[SSWELL_ROGERS, SSWELL_ARDHUIN, SSWELL_ZIEGER],
    Field(description="Swell dissipation component", discriminator="model_type"),
]
WCAPPING_TYPE = Annotated[
    Union[WCAPPING_KOMEN, WCAPPING_AB],
    Field(description="Whitecapping component", discriminator="model_type"),
]
BREAKING_TYPE = Annotated[
    Union[BREAKING_CONSTANT, BREAKING_BKD],
    Field(description="Wave breaking component", discriminator="model_type"),
]
FRICTION_TYPE = Annotated[
    Union[FRICTION_JONSWAP, FRICTION_COLLINS, FRICTION_MADSEN, FRICTION_RIPPLES],
    Field(description="Bottom friction component", discriminator="model_type"),
]
TRIAD_TYPE = Annotated[
    Union[TRIAD, TRIAD_DCTA, TRIAD_LTA, TRIAD_SPB],
    Field(description="Triad interactions component", discriminator="model_type"),
]
SICE_TYPE = Annotated[
    Union[SICE, SICE_R19, SICE_D15, SICE_M18, SICE_R21B],
    Field(description="Sea ice component", discriminator="model_type"),
]
BRAGG_TYPE = Annotated[
    Union[BRAGG, BRAGG_FT, BRAGG_FILE],
    Field(description="Bragg scattering component", discriminator="model_type"),
]


class PHYSICS(BaseGroupComponent):
    """Physics group component.

    The physics group component is a convenience to allow specifying several individual
    components in a single command and check for consistency between them.

    Exemples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.group import PHYSICS
        gen = {"model_type": "gen3", "source_terms": {"model_type": "komen"}}
        phys = PHYSICS(gen=gen)
        print(phys.render())
        phys = PHYSICS(
            gen=dict(model_type="gen3", source_terms={"model_type": "st6c1"}),
            negatinp={"model_type": "negatinp", "rdcoef": 0.04},
            sswell={"model_type": "zieger"},
            breaking={"model_type": "constant", "alpha": 1.0, "gamma": 0.73},
            friction={"model_type": "jonswap", "cfjon": 0.038},
        )
        print(phys.render())

    """

    model_type: Literal["physics", "PHYSICS"] = Field(
        default="physics", description="Model type discriminator"
    )
    gen: Optional[GEN_TYPE] = Field(default=None)
    sswell: Optional[SSWELL_TYPE] = Field(default=None)
    negatinp: Optional[NEGATINP_TYPE] = Field(default=None)
    wcapping: Optional[WCAPPING_TYPE] = Field(default=None)
    quadrupl: Optional[QUADRUPL_TYPE] = Field(default=None)
    breaking: Optional[BREAKING_TYPE] = Field(default=None)
    friction: Optional[FRICTION_TYPE] = Field(default=None)
    triad: Optional[TRIAD_TYPE] = Field(default=None)
    vegetation: Optional[VEGETATION_TYPE] = Field(default=None)
    mud: Optional[MUD_TYPE] = Field(default=None)
    sice: Optional[SICE_TYPE] = Field(default=None)
    turbulence: Optional[TURBULENCE_TYPE] = Field(default=None)
    bragg: Optional[BRAGG_TYPE] = Field(default=None)
    limiter: Optional[LIMITER_TYPE] = Field(default=None)
    obstacle: Optional[OBSTACLE_TYPE] = Field(default=None)
    setup: Optional[SETUP_TYPE] = Field(default=None)
    diffraction: Optional[DIFFRACTION_TYPE] = Field(default=None)
    surfbeat: Optional[SURFBEAT_TYPE] = Field(default=None)
    scat: Optional[SCAT_TYPE] = Field(default=None)
    deactivate: Optional[OFF_TYPE] = Field(default=None)

    @field_validator("deactivate")
    @classmethod
    def deactivate_physics(cls, off: OFF_TYPE) -> OFF_TYPE:
        """Convert OFF to OFFS so list is rendered."""
        for phys in PhysicsOff:
            print(phys.value)
        return off

    @model_validator(mode="after")
    def negatinp_only_with_zieger(self) -> "PHYSICS":
        """Log a warning if NEGATINP is used with a non-ZIEGER SSWELL."""
        if self.negatinp is None:
            return self
        elif self.sswell is None:
            logger.warning(
                "The negative wind input NEGATINP is only intended to use with the "
                "swell dissipation SSWELL ZIEGER but no SSWELL has been specified."
            )
        elif self.sswell.model_type != "zieger":
            logger.warning(
                "The negative wind input NEGATINP is only intended to use with the "
                "swell dissipation SSWELL ZIEGER but the SSWELL "
                f"{self.sswell.model_type.upper()} has been specified."
            )
        return self

    def cmd(self):
        repr = []
        if self.gen is not None:
            repr += [self.gen.cmd()]
        if self.sswell is not None:
            repr += [f"{self.sswell.cmd()}"]
        if self.negatinp is not None:
            repr += [self.negatinp.cmd()]
        if self.wcapping is not None:
            repr += [self.wcapping.cmd()]
        if self.quadrupl is not None:
            repr += [self.quadrupl.cmd()]
        if self.breaking is not None:
            repr += [self.breaking.cmd()]
        if self.friction is not None:
            repr += [self.friction.cmd()]
        if self.triad is not None:
            repr += [self.triad.cmd()]
        if self.vegetation is not None:
            repr += [self.vegetation.cmd()]
        if self.mud is not None:
            repr += [self.mud.cmd()]
        if self.sice is not None:
            repr += [self.sice.cmd()]
        if self.turbulence is not None:
            repr += [self.turbulence.cmd()]
        if self.bragg is not None:
            repr += [self.bragg.cmd()]
        if self.limiter is not None:
            repr += [self.limiter.cmd()]
        if self.obstacle is not None:
            repr += self.obstacle.cmd()  # Object returns a list of components
        if self.setup is not None:
            repr += [self.setup.cmd()]
        if self.diffraction is not None:
            repr += [self.diffraction.cmd()]
        if self.surfbeat is not None:
            repr += [self.surfbeat.cmd()]
        if self.scat is not None:
            repr += [self.scat.cmd()]
        if self.deactivate is not None:
            repr += self.deactivate.cmd()  # Object returns a list of components
        return repr


# =====================================================================================
# Output
# =====================================================================================
FRAME_TYPE = Annotated[FRAME, Field(description="Frame locations component")]
GROUP_TYPE = Annotated[GROUP, Field(description="Group locations component")]
CURVE_TYPE = Annotated[CURVES, Field(description="Curve locations component")]
RAY_TYPE = Annotated[RAY, Field(description="Ray locations component")]
ISOLINE_TYPE = Annotated[ISOLINE, Field(description="Isoline locations component")]
QUANTITY_TYPE = Annotated[QUANTITIES, Field(description="Quantity component")]
OUTOPT_TYPE = Annotated[OUTPUT_OPTIONS, Field(description="Output options component")]
TABLE_TYPE = Annotated[TABLE, Field(description="Table write component")]
SPECOUT_TYPE = Annotated[SPECOUT, Field(description="Spectra write component")]
NESTOUT_TYPE = Annotated[NESTOUT, Field(description="Nest write component")]
TEST_TYPE = Annotated[TEST, Field(description="Intermediate write component")]
BLOCK_TYPE = Annotated[
    Union[BLOCK, BLOCKS],
    Field(description="Block write component", discriminator="model_type"),
]
POINTS_TYPE = Annotated[
    Union[POINTS, POINTS_FILE],
    Field(description="Points locations component", discriminator="model_type"),
]
NGRID_TYPE = Annotated[
    Union[NGRID, NGRID_UNSTRUCTURED],
    Field(description="Ngrid locations component", discriminator="model_type"),
]


class OUTPUT(BaseGroupComponent):
    """Output group component.

    .. code-block:: text

        FRAME 'sname' ...
        GROUP 'sname' ...
        CURVE 'sname' ...
        RAY 'rname' ...
        ISOLINE 'sname' 'rname' ...
        POINTS 'sname ...
        NGRID 'sname' ...
        QUANTITY ...
        OUTPUT OPTIONS ...
        BLOCK 'sname' ...
        TABLE 'sname' ...
        SPECOUT 'sname' ...
        NESTOUT 'sname ...

    This group component is used to define multiple types of output locations and
    write components in a single model. Only fields that are explicitly prescribed are
    rendered by this group component.

    Note
    ----
    The components prescribed are validated according to some constraints as defined
    in the SWAN manual:

    - The name `'sname'` of each Locations component must be unique.
    - The Locations `'sname'` assigned to each write component must be defined.
    - The BLOCK component must be associated with either a `FRAME` or `GROUP`.
    - The ISOLINE write component must be associated with a `RAY` component.
    - The NGRID and NESTOUT components must be defined together.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.output import POINTS, BLOCK, QUANTITIES, TABLE
        from rompy_swan.components.group import OUTPUT
        points = POINTS(sname="outpts", xp=[172.3, 172.4], yp=[-39, -39])
        quantity = QUANTITIES(
            quantities=[
                dict(output=["depth", "hsign", "tps", "dir", "tm01"], excv=-9),
            ]
        )
        times = dict(tbeg="2012-01-01T00:00:00", delt="PT30M", tfmt=1, dfmt="min")
        block = BLOCK(
            model_type="block",
            sname="COMPGRID",
            fname="./swangrid.nc",
            output=["depth", "hsign", "tps", "dir"],
            times=times,
        )
        table = TABLE(
            sname="outpts",
            format="noheader",
            fname="./swantable.nc",
            output=["hsign", "hswell", "dir", "tps", "tm01", "watlev", "qp"],
            times=times,
        )
        out = OUTPUT(
            points=points,
            quantity=quantity,
            block=block,
            table=table,
        )
        print(out.render())

    """

    model_type: Literal["output", "OUTPUT"] = Field(
        default="output", description="Model type discriminator"
    )
    frame: Optional[FRAME_TYPE] = Field(default=None)
    group: Optional[GROUP_TYPE] = Field(default=None)
    curve: Optional[CURVE_TYPE] = Field(default=None)
    ray: Optional[RAY_TYPE] = Field(default=None)
    isoline: Optional[ISOLINE_TYPE] = Field(default=None)
    points: Optional[POINTS_TYPE] = Field(default=None)
    ngrid: Optional[NGRID_TYPE] = Field(default=None)
    quantity: Optional[QUANTITY_TYPE] = Field(default=None)
    output_options: Optional[OUTOPT_TYPE] = Field(default=None)
    block: Optional[BLOCK_TYPE] = Field(default=None)
    table: Optional[TABLE_TYPE] = Field(default=None)
    specout: Optional[SPECOUT_TYPE] = Field(default=None)
    nestout: Optional[NESTOUT_TYPE] = Field(default=None)
    test: Optional[TEST_TYPE] = Field(default=None)
    _location_fields: list = ["frame", "group", "curve", "isoline", "points", "ngrid"]
    _write_fields: list = ["block", "table", "specout", "nestout"]

    @model_validator(mode="after")
    def write_locations_exists(self) -> "OUTPUT":
        """Ensure the location component requested by a write component exists."""
        for write in self.write_set:
            obj = getattr(self, write)
            if obj is None:
                continue
            snames = obj.sname if isinstance(obj.sname, list) else [obj.sname]
            for sname in snames:
                if sname in SPECIAL_NAMES:
                    return self
                try:
                    self._filter_location(sname)
                except ValueError as err:
                    raise ValueError(
                        f"Write component '{write}' specified with sname='{sname}' but "
                        f"no location component with sname='{sname}' has been defined"
                    ) from err
        return self

    @model_validator(mode="after")
    def locations_sname_unique(self) -> "OUTPUT":
        """Ensure same `sname` isn't used in more than one set of output locations."""
        duplicates = {x for x in self.snames if self.snames.count(x) > 1}
        if duplicates:
            raise ValueError(
                "The following snames are used to define more than one set of output "
                f"components: {duplicates}, please ensure each location component has "
                "a unique `sname`"
            )
        return self

    @model_validator(mode="after")
    def block_with_frame_or_group(self) -> "OUTPUT":
        """Ensure Block is only defined for FRAME or GROUP locations."""
        if self.block is not None:
            snames = self.block.sname
            if isinstance(snames, str):
                snames = [self.block.sname]
            for sname in snames:
                if sname not in ["BOTTGRID", "COMPGRID"]:
                    location = self._filter_location(sname)
                    component = location.model_type.upper().split("_")[0]
                    if component not in ["FRAME", "GROUP"]:
                        raise ValueError(
                            f"Block sname='{sname}' specified with {component} "
                            "location component but only only FRAME or GROUP "
                            "components are supported"
                        )
        return self

    @model_validator(mode="after")
    def isoline_ray_defined(self) -> "OUTPUT":
        """Ensure the isoline ray has been defined."""
        if self.isoline is not None:
            if self.ray is None:
                raise ValueError(
                    f"Isoline {self.isoline} requires RAY rname='{self.isoline.rname}'"
                    " but no RAY component has been defined"
                )
            elif self.ray.rname != self.isoline.rname:
                raise ValueError(
                    f"Isoline rname='{self.isoline.rname}' does not match "
                    f"the ray rname='{self.ray.rname}'"
                )
        return self

    @model_validator(mode="after")
    def ngrid_and_nestout(self) -> "OUTPUT":
        """Ensure NGRID and NESTOUT are specified together."""
        if self.ngrid is not None and self.nestout is None:
            raise ValueError(
                "NGRID component specified but no NESTOUT component has been defined"
            )
        elif self.ngrid is None and self.nestout is not None:
            raise ValueError(
                "NESTOUT component specified but no NGRID component has been defined"
            )
        elif self.ngrid is not None and self.nestout is not None:
            if self.ngrid.sname != self.nestout.sname:
                raise ValueError(
                    f"NGRID sname='{self.ngrid.sname}' does not match "
                    f"the NESTOUT sname='{self.nestout.sname}'"
                )
        return self

    @property
    def locations_set(self):
        """List of specified location fields."""
        return [fld for fld in self.model_fields_set if fld in self._location_fields]

    @property
    def write_set(self):
        """List of specified write fields."""
        return [fld for fld in self.model_fields_set if fld in self._write_fields]

    @property
    def snames(self):
        """List of snames from specified location components."""
        snames = []
        for field in self.locations_set:
            obj = getattr(self, field)
            if obj is None:
                continue
            sname = obj.sname
            if isinstance(sname, str):
                sname = [sname]
            snames.extend(sname)
        return snames

    def _filter_location(self, sname):
        """Filter the location component defined with the specified sname."""
        for field in self.locations_set:
            obj = getattr(self, field)
            if obj is None:
                continue
            obj_snames = obj.sname if isinstance(obj.sname, list) else [obj.sname]
            for obj_sname in obj_snames:
                if obj_sname == sname:
                    return obj
        raise ValueError(f"Location component with sname='{sname}' not found")

    def cmd(self) -> list:
        """Command file string for this component."""
        repr = []
        if self.frame is not None:
            repr += [f"{self.frame.cmd()}"]
        if self.group is not None:
            repr += [f"{self.group.cmd()}"]
        if self.curve is not None:
            # Component renders a list
            repr += self.curve.cmd()
        if self.ray is not None:
            repr += [f"{self.ray.cmd()}"]
        if self.isoline is not None:
            repr += [f"{self.isoline.cmd()}"]
        if self.points is not None:
            repr += [f"{self.points.cmd()}"]
        if self.ngrid is not None:
            repr += [f"{self.ngrid.cmd()}"]
        if self.quantity is not None:
            # Component renders a list
            repr += self.quantity.cmd()
        if self.output_options is not None:
            repr += [f"{self.output_options.cmd()}"]
        if self.block is not None:
            # Component may or may not render a list, handles both
            cmds = self.block.cmd()
            if not isinstance(cmds, list):
                cmds = [cmds]
            for cmd in cmds:
                repr += [f"{cmd}"]
        if self.table is not None:
            repr += [f"{self.table.cmd()}"]
        if self.specout is not None:
            repr += [f"{self.specout.cmd()}"]
        if self.nestout is not None:
            repr += [f"{self.nestout.cmd()}"]
        if self.test is not None:
            repr += [f"{self.test.cmd()}"]
        return repr


# =====================================================================================
# Lockup
# =====================================================================================
COMPUTE_TYPE = Annotated[
    Union[COMPUTE_STAT, COMPUTE_NONSTAT],
    Field(description="Compute components", discriminator="model_type"),
]


class LOCKUP(BaseComponent):
    """Lockup group component.

    .. code-block:: text

        COMPUTE ...
        HOTFILE ...
        COMPUTE ...
        HOTFILE ...
        ...
        STOP

    This is a group component to specify SWAN "Lockup" commands including multiple
    `COMPUTE` commands that may or may not be interleaved with `HOTFILE` commands,
    and a final `STOP` command.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.group import LOCKUP
        lockup = LOCKUP(
            compute=dict(
                model_type="stat",
                times=dict(
                    model_type="nonstationary",
                    tbeg="1990-01-01T00:00:00",
                    tend="1990-01-01T03:00:00",
                    delt="PT1H",
                    dfmt="hr",
                ),
                hotfile=dict(fname="hotfile"),
                hottimes=[-1],
            ),
        )
        print(lockup.render())

    """

    model_type: Literal["lockup", "LOCKUP"] = Field(
        default="lockup", description="Model type discriminator"
    )
    compute: COMPUTE_TYPE = Field(description="Compute components")

    def cmd(self) -> list:
        """Command file strings for this component."""
        return self.compute.cmd() + [STOP().render()]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import os
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import rompy_swan

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinxcontrib.autodoc_pydantic",
    "nbsphinx",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "rompy-swan"
copyright = "2021, CSIRO and rompy contributors"
author = "Rompy Developers"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = rompy_swan.__version__
# The full version, including alpha/beta/rc tags.
release = rompy_swan.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#

language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

html_theme_options = {
    "collapse_navigation": True,
    "show_nav_level": 2,
    "secondary_sidebar_items": ["page-toc", "edit-this-page", "sourcelink"],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/rom-py/rompy-swan",
            "icon": "fab fa-github",
            "type": "fontawesome",
        }
    ],
}

html_sidebars = {
    "**": ["globaltoc.html"],
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]
# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "rompy-swan", "Rompy SWAN Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "oceanum",
        "oceanum Documentation",
        author,
        "oceanum",
        "Python library for Oceanum.io.",
        "Miscellaneous",
    ),
]

autoclass_content = "class"

autodoc_pydantic_model_show_json = True
autodoc_pydantic_model_show_config_summary = False
autodoc_pydantic_settings_show_validator_summary = True
man_pages = [(master_doc, "rompy-swan", "Rompy SWAN Documentation", [author], 1)]



"""
SWAN Boundary Subcomponents

This module contains subcomponents for defining boundary conditions in SWAN,
including boundary segments, spectral parameters, and initial conditions.
"""

from typing import Annotated, Literal, Optional, Union

from pydantic import Field, model_validator

from rompy.logging import get_logger
from rompy_swan.subcomponents.base import IJ, XY, BaseSubComponent

logger = get_logger(__name__)


class SIDE(BaseSubComponent):
    """Boundary over one side of computational domain.

    .. code-block:: text

        SIDE NORTH|NW|WEST|SW|SOUTH|SE|E|NE CCW|CLOCKWISE

    The boundary is one full side of the computational grid (in 1D cases either of the
    two ends of the 1D-grid).

    Note
    ----
    Should not be used in case of CURVILINEAR grids.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.boundary import SIDE
        side = SIDE(side="west", direction="ccw")
        print(side.render())

    """

    model_type: Literal["side", "SIDE"] = Field(
        default="side",
        description="Model type discriminator",
    )
    side: Literal["north", "nw", "west", "sw", "south", "se", "east", "ne"] = Field(
        description="The side of the grid to apply the boundary to",
    )
    direction: Literal["ccw", "clockwise"] = Field(
        default="ccw",
        description="The direction to apply the boundary in",
    )

    def cmd(self) -> str:
        repr = f"SIDE {self.side.upper()} {self.direction.upper()} "
        return repr


class SIDES(BaseSubComponent):
    """Boundary over multiple side of computational domain.

    .. code-block:: text

        SIDE NORTH|NW|WEST|SW|SOUTH|SE|E|NE CCW|CLOCKWISE
        SIDE NORTH|NW|WEST|SW|SOUTH|SE|E|NE CCW|CLOCKWISE
        ...

    Note
    ----
    Should not be used in case of CURVILINEAR grids.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.boundary import SIDE, SIDES
        side1 = SIDE(side="west", direction="ccw")
        side2 = SIDE(side="north", direction="ccw")
        sides = SIDES(sides=[side1, side2])
        print(sides.render())

    """

    model_type: Literal["sides", "SIDES"] = Field(
        default="sides",
        description="Model type discriminator",
    )
    sides: list[SIDE] = Field(
        description="The sides of the grid to apply the boundary to",
    )

    def cmd(self) -> str:
        repr = []
        for side in self.sides:
            repr += f"SIDE {side.side.upper()} {side.direction.upper()} "
        return repr


class SEGMENT(BaseSubComponent):
    """Boundary over a segment defined from points.

    .. code-block:: text

        SEGMENT XY < [x] [y] >
        SEGMENT IJ < [i] [j] >

    The segment is defined either by means of a series of points in terms of problem
    coordinates (`XY`) or by means of a series of points in terms of grid indices
    (`IJ`). The points do not have to include all or coincide with actual grid points.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.boundary import SEGMENT
        seg = SEGMENT(
            points=dict(
                model_type="xy",
                x=[172, 172, 172, 172.5, 173],
                y=[-41, -40.5, -40, -40, -40],
                fmt="0.2f",
            ),
        )
        print(seg.render())
        seg = SEGMENT(
            points=dict(
                model_type="ij",
                i=[0, 0, 5],
                j=[0, 19, 19],
            ),
        )
        print(seg.render())

    """

    model_type: Literal["segment", "SEGMENT"] = Field(
        default="segment",
        description="Model type discriminator",
    )
    points: Union[XY, IJ] = Field(
        description="Points to define the segment",
        discriminator="model_type",
    )

    def cmd(self) -> str:
        return f"SEGMENT {self.points.model_type.upper()} {self.points.render()}"


class PAR(BaseSubComponent):
    """Spectral parameters.

    .. code-block:: text

        PAR [hs] [per] [dir] [dd]

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.boundary import PAR
        par = PAR(hs=1.5, per=8.1, dir=225)
        print(par.render())

    """

    model_type: Literal["par"] = Field(
        default="par",
        description="Model type discriminator",
    )
    hs: float = Field(
        description="The significant wave height (m)",
        gt=0.0,
    )
    per: float = Field(
        description=(
            "The characteristic period (s) of the energy spectrum (relative "
            "frequency; which is equal to absolute frequency in the absence of "
            "currents); `per` is the value of the peak period if option PEAK is "
            "chosen in command BOUND SHAPE or `per` is the value of the mean period, "
            "if option MEAN was chosen in command BOUND SHAPE."
        ),
        gt=0.0,
    )
    dir: float = Field(
        description=(
            "The peak wave direction thetapeak (degree), constant over frequencies"
        ),
        ge=-360.0,
        le=360.0,
    )
    dd: Optional[float] = Field(
        default=None,
        description=(
            "Coefficient of directional spreading; a `cos^m(θ)` distribution is "
            "assumed. `dd` is interpreted as the directional standard deviation in "
            "degrees, if the option DEGREES is chosen in the command BOUND SHAPE "
            "(SWAN default: 30). `dd` is interpreted as the power `m`, if the option "
            "POWER is chosen in the command BOUND SHAPE (SWAN default: 2)"
        ),
        ge=0.0,
        le=360.0,
    )

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        repr = f"PAR hs={self.hs} per={self.per} dir={self.dir}"
        if self.dd is not None:
            repr += f" dd={self.dd}"
        return repr


class CONSTANTPAR(PAR):
    """Constant spectral parameters.

    .. code-block:: text

        CONSTANT PAR [hs] [per] [dir] ([dd])

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.boundary import CONSTANTPAR
        par = CONSTANTPAR(hs=1.5, per=8.1, dir=225)
        print(par.render())

    """

    model_type: Literal["constantpar", "CONSTANTPAR"] = Field(
        default="constantpar",
        description="Model type discriminator",
    )

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        return f"CONSTANT {super().cmd()}"


class VARIABLEPAR(BaseSubComponent):
    """Variable spectral parameter.

    .. code-block:: text

        VARIABLE PAR < [len] [hs] [per] [dir] [dd] >

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.boundary import VARIABLEPAR
        par = VARIABLEPAR(
            hs=[1.5, 1.4, 1.1],
            per=[8.1, 8.0, 8.1],
            dir=[225, 226, 228],
            dd=[25, 22, 23],
            len=[0, 0.5, 1.0],
        )
        print(par.render())

    """

    model_type: Literal["variablepar", "VARIABLEPAR"] = Field(
        default="variablepar",
        description="Model type discriminator",
    )
    hs: list[Annotated[float, Field(ge=0.0)]] = Field(
        description="The significant wave height (m)",
    )
    per: list[Annotated[float, Field(ge=0.0)]] = Field(
        description=(
            "The characteristic period (s) of the energy spectrum (relative "
            "frequency; which is equal to absolute frequency in the absence of "
            "currents); `per` is the value of the peak period if option PEAK is "
            "chosen in command BOUND SHAPE or `per` is the value of the mean period, "
            "if option MEAN was chosen in command BOUND SHAPE"
        ),
    )
    dir: list[Annotated[float, Field(ge=-360.0, le=360.0)]] = Field(
        description=(
            "The peak wave direction thetapeak (degrees), constant over frequencies"
        ),
    )
    dd: list[Annotated[float, Field(ge=0.0, le=360.0)]] = Field(
        description=(
            "Coefficient of directional spreading; a $cos^m(θ)$ distribution is "
            "assumed. `dd` is interpreted as the directional standard deviation in "
            "degrees, if the option DEGREES is chosen in the command BOUND SHAPE "
            "(SWAN default: 30). `dd` is interpreted as the power `m`, if the option "
            "POWER is chosen in the command BOUND SHAPE (SWAN default: 2)"
        ),
    )
    dist: list[Annotated[float, Field(ge=0)]] = Field(
        alias="len",
        description=(
            "Is the distance from the first point of the side or segment to the point "
            "along the side or segment for which the incident wave spectrum is "
            "prescribed. Note: these points do no have to coincide with grid points of "
            "the computational grid. `len` is the distance in m or degrees in the case "
            "of spherical coordinates, not in grid steps. The values of `len` should "
            "be given in ascending order. The length along a SIDE is measured in "
            "clockwise or counterclockwise direction, depending on the options CCW or "
            "CLOCKWISE (see above). The option CCW is default. In case of a SEGMENT "
            "the length is measured from the indicated begin point of the segment"
        ),
    )

    @model_validator(mode="after")
    def ensure_equal_size(self) -> "VARIABLEPAR":
        for key in ["hs", "per", "dir", "dd"]:
            if len(getattr(self, key)) != len(self.dist):
                raise ValueError(f"Size of dist and {key} must be the same")
        return self

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        repr = "VARIABLE PAR"
        for dist, hs, per, dir, dd in zip(
            self.dist, self.hs, self.per, self.dir, self.dd
        ):
            repr += f" &\n\tlen={dist} hs={hs} per={per} dir={dir} dd={dd}"
        return repr


class CONSTANTFILE(BaseSubComponent):
    """Constant file specification.

    .. code-block:: text

        CONSTANT FILE 'fname' [seq]

    There are three types of files:

    - TPAR files containing nonstationary wave parameters
    - files containing stationary or nonstationary 1D spectra
      (usually from measurements)
    - files containing stationary or nonstationary 2D spectra
      (from other computer programs or other SWAN runs)

    A TPAR file is for only one location; it has the string TPAR on the first
    line of the file and a number of lines which each contain 5 numbers, i.e.:
    Time (ISO-notation), Hs, Period (average or peak period depending on the
    choice given in command BOUND SHAPE), Peak Direction (Nautical or Cartesian,
    depending on command SET), Directional spread (in degrees or as power of cos
    depending on the choice given in command BOUND SHAPE).

    Note
    ----
    Example of a TPAR file:

    .. code-block:: text

        TPAR
        19920516.130000 4.2 12. -110. 22.
        19920516.180000 4.2 12. -110. 22.
        19920517.000000 1.2 8. -110. 22.
        19920517.120000 1.4 8.5 -80. 26
        19920517.200000 0.9 6.5 -95. 28

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.boundary import CONSTANTFILE
        par = CONSTANTFILE(fname="tpar.txt")
        print(par.render())

    """

    model_type: Literal["constantfile", "CONSTANTFILE"] = Field(
        default="constantfile",
        description="Model type discriminator",
    )
    fname: str = Field(
        description="Name of the file containing the boundary condition.",
        max_length=36,
    )
    seq: Optional[int] = Field(
        default=None,
        description=(
            "sequence number of geographic location in the file (see Appendix D); "
            "useful for files which contain spectra for more than one location. "
            "Note: a TPAR file always contains only one location so in this case "
            "`seq` must always be 1"
        ),
        ge=1,
    )

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        repr = f"CONSTANT FILE fname='{self.fname}'"
        if self.seq is not None:
            repr += f" seq={self.seq}"
        return repr


class VARIABLEFILE(BaseSubComponent):
    """Variable file specification.

    .. code-block:: text

        VARIABLE FILE < [len] 'fname' [seq] >

    There are three types of files:

    - TPAR files containing nonstationary wave parameters
    - files containing stationary or nonstationary 1D spectra
      (usually from measurements)
    - files containing stationary or nonstationary 2D spectra
      (from other computer programs or other SWAN runs)

    A TPAR file is for only one location; it has the string TPAR on the first
    line of the file and a number of lines which each contain 5 numbers, i.e.:
    Time (ISO-notation), Hs, Period (average or peak period depending on the
    choice given in command BOUND SHAPE), Peak Direction (Nautical or Cartesian,
    depending on command SET), Directional spread (in degrees or as power of cos
    depending on the choice given in command BOUND SHAPE).

    Note
    ----
    Example of a TPAR file:

    .. code-block:: text

        TPAR
        19920516.130000 4.2 12. -110. 22.
        19920516.180000 4.2 12. -110. 22.
        19920517.000000 1.2 8. -110. 22.
        19920517.120000 1.4 8.5 -80. 26
        19920517.200000 0.9 6.5 -95. 28

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.boundary import VARIABLEFILE
        par = VARIABLEFILE(
            fname=["tpar1.txt", "tpar2.txt", "tpar3.txt"],
            len=[0.0, 0.5, 1.0],
        )
        print(par.render())

    """

    model_type: Literal["variablefile", "VARIABLEFILE"] = Field(
        default="variablefile",
        description="Model type discriminator",
    )
    fname: list[Annotated[str, Field(max_length=36)]] = Field(
        description="Names of the files containing the boundary condition",
    )
    seq: Optional[list[Annotated[int, Field(ge=1)]]] = Field(
        default=None,
        description=(
            "sequence number of geographic location in the file (see Appendix D); "
            "useful for files which contain spectra for more than one location. "
            "Note: a TPAR file always contains only one location so in this case "
            "[seq] must always be 1."
        ),
    )
    dist: list[Annotated[float, Field(ge=0)]] = Field(
        alias="len",
        description=(
            "Is the distance from the first point of the side or segment to the point "
            "along the side or segment for which the incident wave spectrum is "
            "prescribed. Note: these points do no have to coincide with grid points "
            "of the computational grid. [len] is the distance in m or degrees in the "
            "case of spherical coordinates, not in grid steps. The values of `len` "
            "should be given in ascending order. The length along a SIDE is measured "
            "in clockwise or counterclockwise direction, depending on the options CCW "
            "or CLOCKWISE (see above). The option CCW is default. In case of a "
            "SEGMENT the length is measured from the indicated begin point of the "
            "segment."
        ),
    )

    @model_validator(mode="after")
    def ensure_equal_size(self) -> "VARIABLEFILE":
        for key in ["fname", "seq"]:
            attr = getattr(self, key)
            if attr is not None and len(attr) != len(self.dist):
                raise ValueError(f"Size of dist and {key} must be the same")
        if self.seq is None:
            self.seq = [1] * len(self.dist)
        return self

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        repr = "VARIABLE FILE"
        for dist, fname, seq in zip(self.dist, self.fname, self.seq):
            repr += f" &\n\tlen={dist} fname='{fname}' seq={seq}"
        return repr


class DEFAULT(BaseSubComponent):
    """Default initial conditions.

    .. code-block:: text

        DEFAULT

    The initial spectra are computed from the local wind velocities, using the
    deep-water growth curve of Kahma and Calkoen (1992), cut off at values of
    significant wave height and peak frequency from Pierson and Moskowitz (1964).
    The average (over the model area) spatial step size is used as fetch with local
    wind. The shape of the spectrum is default JONSWAP with a cos2-directional
    distribution (options are available: see command BOUND SHAPE).

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.boundary import DEFAULT
        init = DEFAULT()
        print(init.render())

    """

    model_type: Literal["default", "DEFAULT"] = Field(
        default="default",
        description="Model type discriminator",
    )


class ZERO(BaseSubComponent):
    """Zero initial conditions.

    .. code-block:: text

        ZERO

    The initial spectral densities are all 0; note that if waves are generated in the
    model only by wind, waves can become non-zero only by the presence of the
    ”A” term in the growth model; see the keyword AGROW in command GEN3.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.boundary import ZERO
        init = ZERO()
        print(init.render())

    """

    model_type: Literal["zero", "ZERO"] = Field(
        default="zero",
        description="Model type discriminator",
    )


class HOTSINGLE(BaseSubComponent):
    """Hotstart single initial conditions.

    .. code-block:: text

        HOTSTART SINGLE fname='fname' FREE|UNFORMATTED

    Initial wave field is read from file; this file was generated in a previous SWAN
    run by means of the HOTFILE command. If the previous run was nonstationary,
    the time found on the file will be assumed to be the initial time of computation. It
    can also be used for stationary computation as first guess. The computational grid
    (both in geographical space and in spectral space) must be identical to the one in
    the run in which the initial wave field was computed

    Input will be read from a single (concatenated) hotfile. In the case of a previous
    parallel MPI run, the concatenated hotfile can be created from a set of multiple
    hotfiles using the program hcat.exe, see Implementation Manual.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.boundary import HOTSINGLE
        init = HOTSINGLE(fname="hotstart.swn", format="free")
        print(init.render())

    """

    model_type: Literal["hotsingle", "HOTSINGLE"] = Field(
        default="hotsingle",
        description="Model type discriminator",
    )
    fname: str = Field(
        description="Name of the file containing the initial wave field",
        max_length=36,
    )
    format: Literal["free", "unformatted"] = Field(
        default="free",
        description=(
            "Format of the file containing the initial wave field. "
            "FREE: free format, UNFORMATTED: binary format"
        ),
    )

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        return f"HOTSTART SINGLE fname='{self.fname}' {self.format.upper()}"


class HOTMULTIPLE(BaseSubComponent):
    """Hotstart multiple initial conditions.

    .. code-block:: text

        HOTSTART MULTIPLE fname='fname' FREE|UNFORMATTED

    Initial wave field is read from file; this file was generated in a previous SWAN
    run by means of the HOTFILE command. If the previous run was nonstationary,
    the time found on the file will be assumed to be the initial time of computation. It
    can also be used for stationary computation as first guess. The computational grid
    (both in geographical space and in spectral space) must be identical to the one in
    the run in which the initial wave field was computed

    Input will be read from multiple hotfiles obtained from a previous parallel MPI run.
    The number of files equals the number of processors. Hence, for the present run the
    same number of processors must be chosen.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.boundary import HOTMULTIPLE
        init = HOTMULTIPLE(fname="hotstart.swn", format="free")
        print(init.render())

    """

    model_type: Literal["hotmultiple", "HOTMULTIPLE"] = Field(
        default="hotmultiple",
        description="Model type discriminator",
    )
    fname: str = Field(
        description="Name of the file containing the initial wave field",
        max_length=36,
    )
    format: Literal["free", "unformatted"] = Field(
        default="free",
        description=(
            "Format of the file containing the initial wave field. "
            "FREE: free format, UNFORMATTED: binary format"
        ),
    )

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        return f"HOTSTART MULTIPLE fname='{self.fname}' {self.format.upper()}"



"""
SWAN Boundary Module

This module provides boundary condition handling for the SWAN model within the ROMPY framework.
"""

from abc import ABC
from pathlib import Path
from typing import Literal, Optional, Union

import numpy as np
import pandas as pd
from pydantic import Field, field_validator
from shapely.geometry import LineString

from rompy.core.boundary import BoundaryWaveStation
from rompy.core.time import TimeRange
from rompy.logging import get_logger
from rompy_swan.grid import SwanGrid
from rompy_swan.subcomponents.base import XY
from rompy_swan.subcomponents.boundary import CONSTANTFILE, SEGMENT, SIDE, SIDES
from rompy_swan.subcomponents.spectrum import SHAPESPEC

logger = get_logger(__name__)


def write_tpar(df: pd.DataFrame, filename: str | Path):
    """Write TPAR file.

    Parameters
    ----------
    df : pandas.DataFrame
        TPAR dataframe.
    filename : str | Path
        Filename to write to.

    """
    with open(filename, "w") as stream:
        stream.write("TPAR\n")
        df.to_csv(
            stream,
            sep=" ",
            na_rep=0.0,
            header=False,
            float_format="%0.2f",
            date_format="%Y%m%d.%H%M%S",
        )


class Boundnest1(BoundaryWaveStation):
    """SWAN BOUNDNEST1 NEST data class."""

    model_type: Literal["boundnest1", "BOUNDNEST1"] = Field(
        default="boundnest1", description="Model type discriminator"
    )
    rectangle: Literal["closed", "open"] = Field(
        default="closed",
        description=(
            "Defines whether boundary is defined over an closed or open rectangle"
        ),
    )

    def get(
        self, destdir: str, grid: SwanGrid, time: Optional[TimeRange] = None
    ) -> str:
        """Write the data source to a new location.

        Parameters
        ----------
        destdir : str | Path
            Destination directory for the SWAN ASCII file.
        grid : RegularGrid
            Grid instance to use for selecting the boundary points.
        time: TimeRange, optional
            The times to filter the data to, only used if `self.crop_data` is True.

        Returns
        -------
        filename: Path
            The filename of the written boundary file.
        cmd : str
            Boundary command string to render in the SWAN INPUT file

        """
        if self.crop_data and time is not None:
            self._filter_time(time)
        if self.crop_data and grid is not None:
            self._filter_grid(grid)

        ds = self._sel_boundary(grid).sortby("dir")

        # If nearest, ensure points are returned at the requested positions
        if self.sel_method == "nearest":
            xbnd, ybnd = self._boundary_points(grid=grid)
            ds["lon"].values = xbnd
            ds["lat"].values = ybnd

        filename = Path(destdir) / f"{self.id}.bnd"
        ds.spec.to_swan(filename)
        cmd = f"BOUNDNEST1 NEST '{filename.name}' {self.rectangle.upper()}"
        return filename, cmd


class BoundspecBase(BoundaryWaveStation, ABC):
    """Base class for SWAN BOUNDSPEC data classes."""

    model_type: Literal["boundspecbase", "BOUNDSPECBASE"] = Field(
        default="boundspecbase", description="Model type discriminator"
    )
    shapespec: SHAPESPEC = Field(
        description="Spectral shape specification",
        default=SHAPESPEC(dspr_type="degrees"),
    )
    variable: bool = Field(
        description="Whether the spectra can vary along the side", default=False
    )
    file_type: Literal["tpar", "spec2d"] = Field(
        default="tpar", description="The type of file to write"
    )
    _ds: None

    @field_validator("sel_method_kwargs")
    @classmethod
    def unique_not_supported(cls, kwargs: dict) -> dict:
        if kwargs.get("unique", False):
            raise NotImplementedError(
                "The unique sel option is not support because it will result in a "
                "mismatch between the number of boundary points and the number of "
                "points selected by wavespectra from the dataset."
            )
        return kwargs

    @field_validator("variable")
    @classmethod
    def variable_not_implemented(cls, v, values):
        if v is True:
            raise NotImplementedError("Variable spectra not implemented yet")
        return v

    @property
    def per(self) -> str:
        if self.shapespec.per_type == "peak":
            return "tp"
        elif self.shapespec.per_type == "mean":
            return "tm01"

    @property
    def dspr(self) -> str:
        if self.shapespec.dspr_type == "degrees":
            return "dspr"
        elif self.shapespec.dspr_type == "power":
            raise NotImplementedError("Power of cos not supported yet")

    @property
    def tpar(self) -> pd.DataFrame:
        """TPAR dataframe for the _ds attr."""
        return self._ds.spec.stats(["hs", self.per, "dpm", self.dspr]).to_pandas()

    def _interpolate_side(self, xbnd, ybnd, spacing) -> tuple:
        """Interpolate points along side at user-defined spacing.

        Parameters
        ----------
        xbnd: array-like
            The x coordinates of the boundary points.
        ybnd: array-like
            The y coordinates of the boundary points.
        spacing: float
            The spacing between interpolated points.

        Returns
        -------
        xbnd : array-like
            The x coordinates of the boundary points interpolated at spacing.
        ybnd : array-like
            The y coordinates of the boundary points interpolated at spacing.

        Note
        ----
        The last point is adjusted to ensure the the length of the boundary is not
        exceeded.

        """
        line = LineString(zip(xbnd, ybnd))
        if line.length < spacing:
            raise ValueError(f"Spacing = {spacing} > side length = {line.length}")
        npts = int(np.ceil(line.length / spacing))
        points = [line.interpolate(i * spacing) for i in range(npts + 1)]
        xi = np.array([point.x for point in points])
        yi = np.array([point.y for point in points])
        # Ensure last point does not go beyond the line length
        xi[-1] = xbnd[-1]
        yi[-1] = ybnd[-1]
        return xi, yi

    def _boundary_points_side(self, grid, side):
        """Coordinates of boundary points on a grid side."""
        # Boundary points for grid side in CCW order
        if side.side == "south":
            slc = np.s_[0, :]
        elif side.side == "east":
            slc = np.s_[:, -1]
        elif side.side == "north":
            slc = np.s_[-1, ::-1]
        elif side.side == "west":
            slc = np.s_[::-1, 0]
        elif side.side == "sw":
            slc = np.s_[0, 0]
        elif side.side == "se":
            slc = np.s_[0, -1]
        elif side.side == "nw":
            slc = np.s_[-1, 0]
        elif side.side == "ne":
            slc = np.s_[-1, -1]
        xbnd = grid.x[slc]
        ybnd = grid.y[slc]

        # Reverse if order is clockwise
        if side.direction == "clockwise":
            xbnd = np.flip(xbnd)
            ybnd = np.flip(ybnd)

        # Interpolate at user-defined spacing
        if self.spacing is not None:
            xbnd, ybnd = self._interpolate_side(xbnd, ybnd, self._set_spacing(grid))

        return xbnd, ybnd


class BoundspecSide(BoundspecBase):
    """SWAN BOUNDSPEC SIDE data class.

    TODO: Handle side definition on a rotated grid.
    TODO: Should SIDE VARIABE be supported?
    TODO: Support option to choose between mid-point or averaging?
    TODO: Does PAR need to be supported? Guess not as nonstationary isn't supported

    Note
    ----
    The 'spec1d' file type is not supported yet.

    """

    model_type: Literal["boundspecside", "BOUNDSPECSIDE"] = Field(
        default="boundspecside", description="Model type discriminator"
    )
    location: SIDE = Field(description="The side of the grid to apply the boundary to")

    def _boundary_points(self, grid) -> tuple:
        """Coordinates of boundary points at midpoint of a grid side."""
        xbnd, ybnd = self._boundary_points_side(grid, self.location)
        return [xbnd.mean()], [ybnd.mean()]

    def get(
        self, destdir: str, grid: SwanGrid, time: Optional[TimeRange] = None
    ) -> str:
        """Write the data source to a new location.

        Parameters
        ----------
        destdir : str | Path
            Destination directory for the SWAN ASCII file.
        grid : RegularGrid
            Grid instance to use for selecting the boundary points.
        time: TimeRange, optional
            The times to filter the data to, only used if `self.crop_data` is True.

        Returns
        -------
        filename: Path
            The filename of the written boundary file.
        cmd : str
            Boundary command string to render in the SWAN INPUT file

        """
        if self.crop_data and time is not None:
            self._filter_time(time)
        if self.crop_data and grid is not None:
            self._filter_grid(grid)
        ds = self._sel_boundary(grid).sortby("dir")

        cmds = []
        filenames = []
        # Code below allows for multiple sides but only one side is currently supported
        for ind in range(ds.lon.size):
            self._ds = ds.isel(site=ind, drop=True)
            filename = (
                Path(destdir)
                / f"{self.id}_{self.file_type}_{self.location.side}_{ind:03d}.bnd"
            )
            if self.file_type == "tpar":
                write_tpar(self.tpar, filename)
            elif self.file_type == "spec2d":
                self._ds.spec.to_swan(filename)
            comp = CONSTANTFILE(fname=filename.name, seq=1)
            cmds.append(f"BOUNDSPEC {self.location.render()}{comp.render()}")
            filenames.append(filename)
        return filename, "\n".join(cmds)


class BoundspecSegmentXY(BoundspecBase):
    """SWAN BOUNDSPEC SEGMENT data class.

    TODO: Handle side definition on a rotated grid.
    TODO: Should SIDE VARIABE be supported?
    TODO: Support option to choose between mid-point or averaging?
    TODO: Does PAR need to be supported? Guess not as nonstationary isn't supported
    TODO: If SIDES, ensure continuous

    Note
    ----
    Segments are defined from adjacent point locations so the order in which the points
    are defined is important. When using SIDES, please ensure SIDES are adjacent to
    each other and have correct directions (ccw or clockwise) accordint to the order in
    which each side is prescribed.

    Note
    ----
    The 'spec1d' file type is not supported yet.

    """

    model_type: Literal["boundspecside", "BOUNDSPECSIDE"] = Field(
        default="boundspecside", description="Model type discriminator"
    )
    location: Union[SIDE, SIDES, XY] = Field(
        description="The side of the grid to apply the boundary to",
        discriminator="model_type",
    )

    def _boundary_points(self, grid) -> tuple:
        """Coordinates of boundary points along the segment."""
        if isinstance(self.location, XY):
            xbnd, ybnd = self.location.x, self.location.y
        elif isinstance(self.location, SIDE):
            xbnd, ybnd = self._boundary_points_side(grid, self.location)
        elif isinstance(self.location, SIDES):
            xbnd, ybnd = [], []
            for location in self.location.sides:
                xb, yb = self._boundary_points_side(grid, location)
                # Avoid duplicate points at segment intersection
                if xbnd:
                    xb = xb[1:]
                    yb = yb[1:]
                xbnd.extend(xb)
                ybnd.extend(yb)
        return xbnd, ybnd

    def get(
        self, destdir: str, grid: SwanGrid, time: Optional[TimeRange] = None
    ) -> str:
        """Write the data source to a new location.

        Parameters
        ----------
        destdir : str | Path
            Destination directory for the SWAN ASCII file.
        grid : RegularGrid
            Grid instance to use for selecting the boundary points.
        time: TimeRange, optional
            The times to filter the data to, only used if `self.crop_data` is True.

        Returns
        -------
        filenames: list
            The filenames of the written boundary files.
        cmd : str
            Boundary command string to render in the SWAN INPUT file

        """
        if self.crop_data and time is not None:
            self._filter_time(time)
        if self.crop_data and grid is not None:
            self._filter_grid(grid)
        ds = self._sel_boundary(grid).sortby("dir")

        # If nearest, ensure points are returned at the requested positions
        if self.sel_method == "nearest":
            xbnd, ybnd = self._boundary_points(grid=grid)
            ds["lon"].values = xbnd
            ds["lat"].values = ybnd

        cmds = []
        filenames = []
        for ind in range(ds.lon.size - 1):
            ds_seg = ds.isel(site=slice(ind, ind + 2))
            # TODO: Ensure points in segment are different
            self._ds = ds_seg.mean("site")
            filename = Path(destdir) / f"{self.id}_{self.file_type}_{ind:03d}.bnd"
            if self.file_type == "tpar":
                write_tpar(self.tpar, filename)
            elif self.file_type == "spec2d":
                self._ds.spec.to_swan(filename)
            file = CONSTANTFILE(fname=filename.name, seq=1)
            location = SEGMENT(points=XY(x=ds_seg.lon.values, y=ds_seg.lat.values))
            location = location.render().replace("\n", " ").replace("  ", " ")
            cmds.append(f"BOUNDSPEC {location}{file.render()}")
            filenames.append(filename)
        return filenames, "\n".join(cmds)



"""Test base component."""

from string import ascii_lowercase, ascii_uppercase
from typing import Literal

# Import test utilities
from test_utils.logging import get_test_logger

# Initialize logger
logger = get_test_logger(__name__)

from rompy_swan.components.base import MAX_LENGTH, BaseComponent


class LongRender(BaseComponent):
    """Long render method."""

    model_type: Literal["test"] = "test"

    @property
    def options(self):
        return {c: f"{c}_value" for c in ascii_lowercase + ascii_uppercase}

    def cmd(self):
        """Render the component to a string."""
        repr = "COMPONENT"
        for k, v in self.options.items():
            if k == "A":
                repr += f"\n{k}={v}"
            else:
                repr += f" {k}={v}"
        return repr


def test_max_line_size():
    """Test that the render string is not longer than 180 characters."""
    lr = LongRender()
    for cmd_line in lr.render().split("\n"):
        assert len(cmd_line) <= MAX_LENGTH



"""
SWAN Read Grid Subcomponents

This module contains subcomponents for reading grid data in SWAN,
including regular grids and various input formats.
"""

from abc import ABC
from typing import Literal, Optional, Union

from pydantic import Field, field_validator, model_validator

from rompy.logging import get_logger
from rompy_swan.subcomponents.base import BaseSubComponent
from rompy_swan.types import IDLA, GridOptions

logger = get_logger(__name__)


class GRIDREGULAR(BaseSubComponent):
    """SWAN Regular Grid subcomponent.

    .. code-block:: text

        xp yp alp xlen ylen mx my

    Note
    ----
    The direction of the x-axis `alp` must be 0 in case of spherical coordinates

    Note
    ----
    All coordinates and distances should be given in m when Cartesian coordinates are
    used or degrees when Spherical coordinates are used (see command COORD).

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.readgrid import GRIDREGULAR
        kwargs = dict(xp=173, yp=-40, alp=0, xlen=2, ylen=2, mx=199, my=199)
        grid = GRIDREGULAR(suffix="c", **kwargs)
        print(grid.render())
        grid = GRIDREGULAR(suffix="inp", **kwargs)
        print(grid.render())

    """

    model_type: Literal["gridregular", "GRIDREGULAR"] = Field(
        default="gridregular", description="Model type discriminator"
    )
    xp: float = Field(
        description="The x-coordinate of the origin in problem coordinates",
    )
    yp: float = Field(
        description="The y-coordinate of the origin in problem coordinates",
    )
    alp: Optional[float] = Field(
        default=0.0,
        description="Direction of the xaxis in degrees",
    )
    xlen: float = Field(
        description="Length of the computational grid in the x-direction"
    )
    ylen: float = Field(
        description="Length of the computational grid in the y-direction"
    )
    mx: int = Field(
        description=(
            "Number of meshes in computational grid in x-direction (this number is "
            "one less than the number of grid points in this domain)"
        ),
    )
    my: int = Field(
        description=(
            "Number of meshes in computational grid in y-direction (this number is "
            "one less than the number of grid points in this domain)"
        ),
    )
    suffix: Optional[str] = Field(
        default="", description="Suffix for rendering with each output grid parameter."
    )

    @property
    def dx(self):
        """Grid spacing in x-direction."""
        return self.xlen / self.mx

    @property
    def dy(self):
        """Grid spacing in y-direction."""
        return self.ylen / self.my

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        repr = f"xp{self.suffix}={self.xp}"
        repr += f" yp{self.suffix}={self.yp}"
        repr += f" alp{self.suffix}={self.alp}"
        repr += f" xlen{self.suffix}={self.xlen}"
        repr += f" ylen{self.suffix}={self.ylen}"
        repr += f" mx{self.suffix}={self.mx}"
        repr += f" my{self.suffix}={self.my}"
        return repr


class READGRID(BaseSubComponent, ABC):
    """SWAN grid reader abstract class.

    .. code-block:: text

        READGRID [grid_type] [fac] 'fname1' [idla] [nhedf] ([nhedt]) ([nhedvec]) &
            ->FREE|FORMAT|UNFORMATTED ('form'|[idfm])

    This is the base class for all input grids. It is not meant to be used directly.

    Note
    ----

    File format identifier:

    * 1: Format according to BODKAR convention (a standard of the Ministry of
      Transport and Public Works in the Netherlands). Format string: (10X,12F5.0)
    * 5: Format (16F5.0), an input line consists of 16 fields of 5 places each
    * 6: Format (12F6.0), an input line consists of 12 fields of 6 places each
    * 8: Format (10F8.0), an input line consists of 10 fields of 8 places each

    """

    model_type: Literal["readgrid", "READGRID"] = Field(
        default="readgrid", description="Model type discriminator"
    )
    grid_type: Union[GridOptions, Literal["coordinates"]] = Field(
        description="Type of the SWAN grid file",
    )
    fac: float = Field(
        default=1.0,
        description=(
            "SWAN multiplies all values that are read from file by `fac`. For instance "
            "if the values are given in unit decimeter, one should make `fac=0.1` to "
            "obtain values in m. To change sign use a negative `fac`"
        ),
        gt=0.0,
    )
    idla: IDLA = Field(
        default=1,
        description=(
            "Prescribes the order in which the values of bottom levels "
            "and other fields should be given in the file"
        ),
    )
    nhedf: int = Field(
        default=0,
        description=(
            "The number of header lines at the start of the file. The text in the "
            "header lines is reproduced in the print file created by SWAN . The file "
            "may start with more header lines than `nhedf` because the start of the "
            "file is often also the start of a time step and possibly also of a "
            "vector variable (each having header lines, see `nhedt` and `nhedvec`)"
        ),
        ge=0,
    )
    nhedvec: int = Field(
        default=0,
        description=(
            "For each vector variable: number of header lines in the file "
            "at the start of each component (e.g., x- or y-component)"
        ),
        ge=0,
    )
    format: Literal["free", "fixed", "unformatted"] = Field(
        default="free",
        description=(
            "File format, one of 'free', 'fixed' or 'unformatted'. If 'free', the "
            "file is assumed to use the FREE FORTRAN format. If 'fixed', the file is "
            "assumed to use a fixed format that must be specified by (only) one of "
            "'form' or 'idfm' arguments. Use 'unformatted' to read unformatted "
            "(binary) files (not recommended for ordinary use)"
        ),
    )
    form: Optional[str] = Field(
        default=None,
        description=(
            "A user-specified format string in Fortran convention, e.g., '(10X,12F5.0)'."
            "Only used if `format='fixed'`, do not use it if `idfm` is specified"
        ),
    )
    idfm: Optional[Literal[1, 5, 6, 8]] = Field(
        default=None,
        description=("File format identifier, only used if `format='fixed'`"),
    )

    @model_validator(mode="after")
    def check_format_definition(self) -> "READGRID":
        """Check the arguments specifying the file format are specified correctly."""
        if self.format == "free" and any([self.form, self.idfm]):
            logger.warn(f"FREE format, ignoring form={self.form} idfm={self.idfm}")
        elif self.format == "unformatted" and any([self.form, self.idfm]):
            logger.warn(
                f"UNFORMATTED format, ignoring form={self.form} idfm={self.idfm}"
            )
        elif self.format == "fixed" and not any([self.form, self.idfm]):
            raise ValueError(
                "FIXED format requires one of form or idfm to be specified"
            )
        elif self.format == "fixed" and all([self.form, self.idfm]):
            raise ValueError("FIXED format accepts only one of form or idfm")
        return self

    @property
    def format_repr(self):
        if self.format == "free":
            repr = "FREE"
        elif self.format == "fixed" and self.form:
            repr = f"FORMAT form='{self.form}'"
        elif self.format == "fixed" and self.idfm:
            repr = f"FORMAT idfm={self.idfm}"
        elif self.format == "unformatted":
            repr = "UNFORMATTED"
        return repr


class READCOORD(READGRID):
    """SWAN coordinates reader.

    .. code-block:: text

        READGRID COORDINATES [fac] 'fname' [idla] [nhedf] [nhedvec] &
            FREE|FORMAT ('form'|idfm)

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.readgrid import READCOORD
        readcoord = READCOORD(
            fac=1.0,
            fname="coords.txt",
            idla=3,
            format="free",
        )
        print(readcoord.render())

    """

    model_type: Literal["readcoord", "READCOORD"] = Field(
        default="readcoord", description="Model type discriminator"
    )
    grid_type: Literal["coordinates"] = Field(
        default="coordinates", description="Type of the SWAN grid file"
    )
    fname: str = Field(description="Name of the SWAN coordinates file")

    def cmd(self) -> str:
        repr = (
            f"READGRID COORDINATES fac={self.fac} fname='{self.fname}' "
            f"idla={self.idla} nhedf={self.nhedf} nhedvec={self.nhedvec} "
            f"{self.format_repr}"
        )
        return repr


class READINP(READGRID):
    """SWAN input grid reader.

    .. code-block:: text

        READINP GRID_TYPE [fac] ('fname1' | SERIES 'fname2') [idla] [nhedf] &
            ([nhedt]) [nhedvec] FREE|FORMAT ('form'|idfm)|UNFORMATTED`

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.readgrid import READINP
        readinp = READINP(
            grid_type="wind",
            fname1="wind.txt",
            fac=1.0,
            idla=3,
            format="free",
        )
        print(readinp.render())

    """

    model_type: Literal["readinp", "READINP"] = Field(
        default="readinp", description="Model type discriminator"
    )
    grid_type: Optional[GridOptions] = Field(
        default=None, description="Type of the SWAN grid file"
    )
    fname1: str = Field(
        description="Name of the file with the values of the variable.",
    )
    fname2: Optional[str] = Field(
        default=None,
        description=(
            "Name of file that contains the names of the files where the variables "
            "are given when the SERIES option is used. These names are to be given in "
            "proper time sequence. SWAN reads the next file when the previous file "
            "end has been encountered. In these files the input should be given in "
            "the same format as in the above file 'fname1' (that implies that a file "
            "should start with the start of an input time step)"
        ),
    )
    nhedt: int = Field(
        default=0,
        description=(
            "Only if variable is time dependent: number of header lines in the file "
            "at the start of each time level. A time step may start with more header "
            "lines than `nhedt` because the variable may be a vector variable which "
            "has its own header lines (see `nhedvec`)"
        ),
        ge=0,
    )

    @field_validator("grid_type")
    @classmethod
    def set_undefined(cls, v: str | None) -> str:
        """Allow for undefined value so it can be redefined in INPGRID components."""
        if v is None:
            return "undefined"
        return v

    def cmd(self) -> str:
        repr = f"READINP {self.grid_type.upper()} fac={self.fac} fname1='{self.fname1}'"
        if self.fname2:
            repr += f" SERIES fname2='{self.fname2}'"
        repr += f" idla={self.idla} nhedf={self.nhedf} nhedt={self.nhedt} nhedvec={self.nhedvec} {self.format_repr}"
        return repr



"""
SWAN Time Subcomponents

This module contains subcomponents for handling time specifications in SWAN,
including time ranges, intervals, and time format conversions.
"""

from datetime import datetime, timedelta
from typing import Literal, Union

import pandas as pd
from pydantic import Field, field_validator

from rompy.logging import get_logger
from rompy_swan.subcomponents.base import BaseSubComponent

logger = get_logger(__name__)

DEFAULT_TIME = datetime(1970, 1, 1, 0, 0, 0)
DEFAULT_TEND = DEFAULT_TIME + timedelta(days=1)
DEFAULT_DELT = timedelta(hours=1)
TIME_FORMAT = {
    1: "%Y%m%d.%H%M%S",
    2: "'%d-%b-%y %H:%M:%S'",
    3: "%m/%d/%y.%H:%M:%S",
    4: "%H:%M:%S",
    5: "%y/%m/%d %H:%M:%S'",
    6: "%y%m%d%H%M",
}


class Time(BaseSubComponent):
    """Time specification in SWAN.

    .. code-block:: text

        [time]

    Time is rendered in one of the following formats:

    * 1: ISO-notation 19870530.153000
    * 2: (as in HP compiler) '30-May-87 15:30:00'
    * 3: (as in Lahey compiler) 05/30/87.15:30:00
    * 4: 15:30:00
    * 5: 87/05/30 15:30:00'
    * 6: as in WAM 8705301530

    Note
    ----
    The `time` field can be specified as:

    * existing datetime object
    * int or float, assumed as Unix time, i.e. seconds (if >= -2e10 or <= 2e10) or
      milliseconds (if < -2e10 or > 2e10) since 1 January 1970.
    * ISO 8601 time string.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.time import Time
        from datetime import datetime
        time = Time(time=datetime(1990, 1, 1))
        print(time.render())
        time = Time(time="2012-01-01T00:00:00", tfmt=2)
        print(time.render())

    """

    model_type: Literal["time", "Time", "TIME"] = Field(
        default="time", description="Model type discriminator"
    )
    time: datetime = Field(description="Datetime specification")
    tfmt: Union[Literal[1, 2, 3, 4, 5, 6], str] = Field(
        default=1,
        description="Format to render time specification",
        validate_default=True,
    )

    @field_validator("tfmt")
    @classmethod
    def set_time_format(cls, v: int | str) -> str:
        """Set the time format to render."""
        if isinstance(v, str):
            return v
        return TIME_FORMAT[v]

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        return f"{self.time.strftime(self.tfmt)}"


class Delt(BaseSubComponent):
    """Time interval specification in SWAN.

    .. code-block:: text

        [delt] SEC|MIN|HR|DAY

    Note
    ----
    The `tdelta` field can be specified as:

    * existing timedelta object
    * int or float, assumed as seconds
    * ISO 8601 duration string, following formats work:

      * `[-][DD ][HH:MM]SS[.ffffff]`
      * `[±]P[DD]DT[HH]H[MM]M[SS]S` (ISO 8601 format for timedelta)

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.time import Delt
        from datetime import timedelta
        delt = Delt(delt=timedelta(minutes=30))
        print(delt.render())
        delt = Delt(delt="PT1H", dfmt="hr")
        print(delt.render())

    """

    model_type: Literal["delt"] = Field(
        default="delt", description="Model type discriminator"
    )
    delt: timedelta = Field(description="Time interval")
    dfmt: Literal["sec", "min", "hr", "day"] = Field(
        default="sec",
        description="Format to render time interval specification",
    )

    @property
    def delt_float(self):
        delt_scaling = {"sec": 1, "min": 60, "hr": 3600, "day": 86400}
        return self.delt.total_seconds() / delt_scaling[self.dfmt]

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        return f"{self.delt_float} {self.dfmt.upper()}"


class TimeRangeOpen(BaseSubComponent):
    """Regular times with an open boundary.

    .. code-block:: text

        [tbeg] [delt] SEC|MIN|HR|DAY

    Time is rendered in one of the following formats:

    * 1: ISO-notation 19870530.153000
    * 2: (as in HP compiler) '30-May-87 15:30:00'
    * 3: (as in Lahey compiler) 05/30/87.15:30:00
    * 4: 15:30:00
    * 5: 87/05/30 15:30:00'
    * 6: as in WAM 8705301530

    Note
    ----
    The `tbeg` field can be specified as:

    * existing datetime object
    * int or float, assumed as Unix time, i.e. seconds (if >= -2e10 or <= 2e10) or
      milliseconds (if < -2e10 or > 2e10) since 1 January 1970.
    * ISO 8601 time string.

    Note
    ----
    The `tdelta` field can be specified as:

    * existing timedelta object
    * int or float, assumed as seconds
    * ISO 8601 duration string, following formats work:

        * `[-][DD ][HH:MM]SS[.ffffff]`
        * `[±]P[DD]DT[HH]H[MM]M[SS]S` (ISO 8601 format for timedelta)

    Note
    ----
    Default values for the time specification fields are provided for the case where
    the user wants to set times dynamically after instantiating this subcomponent.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.time import TimeRangeOpen
        from datetime import datetime, timedelta
        times = TimeRangeOpen(
            tbeg=datetime(1990, 1, 1), delt=timedelta(minutes=30), dfmt="min"
        )
        print(times.render())
        times = TimeRangeOpen(
            tbeg="2012-01-01T00:00:00", delt="PT1H", tfmt=2, dfmt="hr", suffix="blk"
        )
        print(times.render())

    """

    model_type: Literal["open", "OPEN"] = Field(
        default="open", description="Model type discriminator"
    )
    tbeg: datetime = Field(default=DEFAULT_TIME, description="Start time")
    delt: timedelta = Field(default=DEFAULT_DELT, description="Time interval")
    tfmt: Union[Literal[1, 2, 3, 4, 5, 6], str] = Field(
        default=1,
        description="Format to render time specification",
    )
    dfmt: Literal["sec", "min", "hr", "day"] = Field(
        default="sec",
        description="Format to render time interval specification",
    )
    suffix: str = Field(
        default="",
        description="Suffix to prepend to argument names when rendering",
    )

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        repr = f"tbeg{self.suffix}={Time(time=self.tbeg, tfmt=self.tfmt).render()}"
        repr += f" delt{self.suffix}={Delt(delt=self.delt, dfmt=self.dfmt).render()}"
        return repr


class TimeRangeClosed(TimeRangeOpen):
    """Regular times with a closed boundary.

    .. code-block:: text

        [tbeg] [delt] SEC|MIN|HR|DAY [tend]

    Note
    ----
    Default values for the time specification fields are provided for the case where
    the user wants to set times dynamically after instantiating this subcomponent.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.time import TimeRangeClosed
        from datetime import datetime, timedelta
        times = TimeRangeClosed(
            tbeg=datetime(1990, 1, 1),
            tend=datetime(1990, 1, 7),
            delt=timedelta(minutes=30),
            dfmt="min",
        )
        print(times.render())
        times = TimeRangeClosed(
            tbeg="2012-01-01T00:00:00",
            tend="2012-02-01T00:00:00",
            delt="PT1H",
            tfmt=2,
            dfmt="hr",
            suffix="blk",
        )
        print(times.render())

    """

    model_type: Literal["closed", "CLOSED"] = Field(
        default="closed", description="Model type discriminator"
    )
    tend: datetime = Field(default=DEFAULT_TEND, description="End time")

    def __call__(self) -> list[Time]:
        """Returns the list of Time objects."""
        times = pd.date_range(start=self.tbeg, end=self.tend, freq=self.delt)
        return [time.to_pydatetime() for time in times]

    def __getitem__(self, index) -> datetime | list[datetime]:
        """Slicing from the times array."""
        return self.__call__()[index]

    def __len__(self):
        """Returns the length of the times array."""
        return len(self())

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        repr = super().cmd()
        repr += f" tend{self.suffix}={Time(time=self.tend, tfmt=self.tfmt).render()}"
        return repr


class NONSTATIONARY(TimeRangeClosed):
    """Nonstationary time specification.

    .. code-block:: text

        NONSTATIONARY [tbeg] [delt] SEC|MIN|HR|DAY [tend]

    Note
    ----
    Default values for the time specification fields are provided for the case where
    the user wants to set times dynamically after instantiating this subcomponent.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.time import NONSTATIONARY
        nonstat = NONSTATIONARY(
            tbeg="2012-01-01T00:00:00",
            tend="2012-02-01T00:00:00",
            delt="PT1H",
            dfmt="hr",
        )
        print(nonstat.render())
        from datetime import datetime, timedelta
        nonstat = NONSTATIONARY(
            tbeg=datetime(1990, 1, 1),
            tend=datetime(1990, 1, 7),
            delt=timedelta(minutes=30),
            tfmt=1,
            dfmt="min",
            suffix="tbl",
        )
        print(nonstat.render())

    """

    model_type: Literal["nonstationary", "NONSTATIONARY"] = Field(
        default="nonstationary", description="Model type discriminator"
    )
    tbeg: datetime = Field(default=DEFAULT_TIME, description="Start time")
    tend: datetime = Field(default=DEFAULT_TEND, description="End time")
    delt: timedelta = Field(default=DEFAULT_DELT, description="Time interval")

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        repr = f"NONSTATIONARY {super().cmd()}"
        return repr


class STATIONARY(BaseSubComponent):
    """Stationary time specification.

    .. code-block:: text

        STATIONARY [time]

    Note
    ----
    The field `time` is optional to allow for the case where the user wants to set the
    time dynamically after instantiating this component.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.time import STATIONARY
        stat = STATIONARY(time="2012-01-01T00:00:00")
        print(stat.render())

    """

    model_type: Literal["stationary", "STATIONARY"] = Field(
        default="stationary", description="Model type discriminator"
    )
    time: datetime = Field(default=DEFAULT_TIME, description="Stationary time")
    tfmt: Union[Literal[1, 2, 3, 4, 5, 6], str] = Field(
        default=1,
        description="Format to render time specification",
    )

    def __call__(self) -> list[Time]:
        """Returns the list of Time object for consistency with NONSTATIONARY."""
        return [self.time]

    def __getitem__(self, index) -> Time | list[Time]:
        """Slicing from the times array."""
        return self.__call__()[index]

    def __len__(self):
        """Returns the length of the times array."""
        return len(self())

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        return f"STATIONARY time={Time(time=self.time, tfmt=self.tfmt).render()}"



"""Test inpgrid component."""

import pytest
from pydantic import ValidationError

# Import test utilities
from test_utils.logging import get_test_logger

# Initialize logger
logger = get_test_logger(__name__)

from rompy_swan.components.group import INPGRIDS
from rompy_swan.components.inpgrid import (
    CURVILINEAR,
    INPGRID,
    READINP,
    REGULAR,
    UNSTRUCTURED,
)
from rompy_swan.subcomponents.time import NONSTATIONARY
from rompy_swan.types import GridOptions


@pytest.fixture(scope="module")
def readinp():
    yield READINP(fname1="test.txt")


@pytest.fixture(scope="module")
def nonstat():
    yield NONSTATIONARY(
        tbeg="2023-01-01T00:00:00",
        tend="2023-02-01T00:00:00",
        delt="PT30M",
    )


def test_valid_inpgrid_options(readinp):
    for inpgrid in GridOptions:
        INPGRID(grid_type=inpgrid.value, readinp=readinp)
    with pytest.raises(ValidationError):
        INPGRID(grid_type="invalid", readinp=readinp)


def test_excval(readinp):
    inpgrid = INPGRID(grid_type="bottom", excval=-999, readinp=readinp)
    assert inpgrid.excval == -999.0
    assert isinstance(inpgrid.excval, float)


def test_inpgrid_nonstationary(nonstat, readinp):
    inpgrid = INPGRID(grid_type="bottom", nonstationary=nonstat, readinp=readinp)
    assert isinstance(inpgrid.nonstationary, NONSTATIONARY)
    assert inpgrid.nonstationary.suffix == "inp"


def test_inpgrid_regular(nonstat, readinp):
    inpgrid = REGULAR(
        grid_type=GridOptions.WIND,
        xpinp=0.0,
        ypinp=0.0,
        alpinp=0.0,
        mxinp=10,
        myinp=10,
        dxinp=0.1,
        dyinp=0.1,
        excval=-999.0,
        nonstationary=nonstat,
        readinp=readinp,
    )
    logger.info(inpgrid.render())


def test_inpgrid_curvilinear(nonstat, readinp):
    inpgrid = CURVILINEAR(
        grid_type=GridOptions.WIND,
        mxinp=10,
        myinp=10,
        excval=-999.0,
        nonstationary=nonstat,
        readinp=readinp,
    )
    logger.info(inpgrid.render())


def test_inpgrid_curvilinear_render(nonstat, readinp):
    inpgrid = UNSTRUCTURED(
        grid_type=GridOptions.WIND,
        excval=-999.0,
        nonstationary=nonstat,
        readinp=readinp,
    )
    logger.info(inpgrid.render())


def test_inpgrids(nonstat, readinp):
    bottom = REGULAR(
        grid_type=GridOptions.BOTTOM,
        xpinp=0.0,
        ypinp=0.0,
        alpinp=0.0,
        mxinp=10,
        myinp=10,
        dxinp=0.1,
        dyinp=0.1,
        excval=-999.0,
        readinp=readinp,
    )
    wind = REGULAR(
        grid_type=GridOptions.WIND,
        xpinp=0.0,
        ypinp=0.0,
        alpinp=0.0,
        mxinp=10,
        myinp=10,
        dxinp=0.1,
        dyinp=0.1,
        excval=-999.0,
        nonstationary=nonstat,
        readinp=readinp,
    )

    inpgrids = INPGRIDS(inpgrids=[bottom, wind])
    logger.info(inpgrids.render())


def test_inpgrids_unique_var(nonstat, readinp):
    bottom = REGULAR(
        grid_type=GridOptions.BOTTOM,
        xpinp=0.0,
        ypinp=0.0,
        alpinp=0.0,
        mxinp=10,
        myinp=10,
        dxinp=0.1,
        dyinp=0.1,
        excval=-999.0,
        readinp=readinp,
    )
    wind = REGULAR(
        grid_type=GridOptions.BOTTOM,
        xpinp=0.0,
        ypinp=0.0,
        alpinp=0.0,
        mxinp=10,
        myinp=10,
        dxinp=0.1,
        dyinp=0.1,
        excval=-999.0,
        nonstationary=nonstat,
        readinp=readinp,
    )

    with pytest.raises(ValidationError):
        INPGRIDS(inpgrids=[bottom, wind])



"""Test swan_config class."""

import os
from pathlib import Path

import pytest
from envyaml import EnvYAML

# Import test utilities
from test_utils.logging import get_test_logger

from rompy.core.source import SourceIntake
from rompy.model import ModelRun
from rompy_swan import Boundnest1
from rompy_swan.config import SwanConfigComponents
from rompy_swan.interface import BoundaryInterface

# Initialize logger
logger = get_test_logger(__name__)

HERE = Path(__file__).parent

os.environ["ROMPY_PATH"] = str(HERE.parent)


@pytest.fixture(scope="module")
def config_dict():
    yield EnvYAML(HERE / "swan_model.yml")


def test_swan_model(tmpdir, config_dict):
    config = SwanConfigComponents(
        startup=config_dict["startup"],
        cgrid=config_dict["cgrid"],
        inpgrid=config_dict["inpgrid"],
        boundary=config_dict["boundary"],
        initial=config_dict["initial"],
        physics=config_dict["physics"],
        prop=config_dict["prop"],
        numeric=config_dict["numeric"],
        output=config_dict["output"],
        lockup=config_dict["lockup"],
    )
    model = ModelRun(
        run_id="test",
        period=dict(start="20230101T00", duration="12h", interval="1h"),
        output_dir=str(tmpdir),
        config=config,
    )
    model.generate()


def test_swan_model_boundary(tmpdir, config_dict):
    config = SwanConfigComponents(
        startup=config_dict["startup"],
        cgrid=config_dict["cgrid"],
        inpgrid=config_dict["inpgrid"],
        boundary=BoundaryInterface(
            kind=Boundnest1(
                id="wave_forcing",
                source=SourceIntake(
                    dataset_id="ausspec", catalog_uri=HERE / "data/catalog.yaml"
                ),
            )
        ),
        initial=config_dict["initial"],
        physics=config_dict["physics"],
        prop=config_dict["prop"],
        numeric=config_dict["numeric"],
        output=config_dict["output"],
        lockup=config_dict["lockup"],
    )
    model = ModelRun(
        run_id="test",
        period=dict(start="20230101T00", duration="12h", interval="1h"),
        output_dir=str(tmpdir),
        config=config,
    )
    model.generate()



import oceantide
from dask.diagnostics.progress import ProgressBar
from oceanum.datamesh import Connector

dataset_id = "tpxo9_v5a_cons"  # ~16km tpxo data
# dataset_id = "??"  # atlas
# dataset_id = "oceanum_global_tide_cons_v2" # 400m downscaled tpxo data

datamesh = Connector()
dset = datamesh.load_datasource(dataset_id)
ds = dset.sel(lon=slice(145, 154.5), lat=slice(-25, -16))
with ProgressBar():
    ds = ds.compute()
ds.tide.to_otis_netcdf("./")



"""Test output components."""

import copy

import pytest

# Import test utilities
from test_utils.logging import get_test_logger

# Initialize logger
logger = get_test_logger(__name__)

import numpy as np
from pydantic import ValidationError

from rompy_swan.components.group import OUTPUT
from rompy_swan.components.output import (
    BLOCK,
    BLOCKS,
    CURVE,
    CURVES,
    FRAME,
    GROUP,
    ISOLINE,
    NESTOUT,
    NGRID,
    NGRID_UNSTRUCTURED,
    OUTPUT_OPTIONS,
    POINTS,
    POINTS_FILE,
    QUANTITIES,
    QUANTITY,
    RAY,
    SPECIAL_NAMES,
    SPECOUT,
    TABLE,
    TEST,
    BaseLocation,
)
from rompy_swan.subcomponents.time import TimeRangeOpen


@pytest.fixture(scope="module")
def times():
    yield TimeRangeOpen(tbeg="1990-01-01T00:00:00", delt="PT1H", tfmt=1, dfmt="hr")


@pytest.fixture(scope="module")
def frame():
    yield FRAME(
        sname="outgrid",
        grid=dict(xp=173, yp=-40, xlen=2, ylen=2, mx=19, my=19),
    )


@pytest.fixture(scope="module")
def group():
    yield GROUP(sname="subgrid", ix1=20, iy1=0, ix2=50, iy2=100)


@pytest.fixture(scope="module")
def curve():
    yield CURVE(
        sname="curve1",
        xp1=172,
        yp1=-40,
        npts=[3, 3],
        xp=[172.0, 174.0],
        yp=[-38.0, -38.0],
    )


@pytest.fixture(scope="module")
def curves(curve):
    curve2 = copy.deepcopy(curve)
    curve2.sname = "curve2"
    yield CURVES(curves=[curve, curve2])


@pytest.fixture(scope="module")
def ray():
    yield RAY(
        rname="outray",
        xp1=171.9,
        yp1=-40.1,
        xq1=172.1,
        yq1=-39.9,
        npts=[3, 3],
        xp=[171.9, 173.9],
        yp=[-38.1, -38.1],
        xq=[172.1, 174.1],
        yq=[-37.9, -37.9],
    )


@pytest.fixture(scope="module")
def isoline():
    yield ISOLINE(sname="outcurve", rname="outray", dep_type="depth", dep=12.0)


@pytest.fixture(scope="module")
def points():
    yield POINTS(sname="outpts", xp=[172.3, 172.4], yp=[-39, -39])


@pytest.fixture(scope="module")
def ngrid():
    yield NGRID(
        sname="outnest", grid=dict(xp=173, yp=-40, xlen=2, ylen=2, mx=19, my=19)
    )


@pytest.fixture(scope="module")
def quantity():
    yield QUANTITY(output=["hsign", "tm02", "fspr"], fmin=0.03, fmax=0.5)


@pytest.fixture(scope="module")
def quantities():
    yield QUANTITIES(
        quantities=[
            QUANTITY(output=["hsign", "tm02", "fspr"], fmin=0.03, fmax=0.5),
            QUANTITY(output=["hswell"], fswell=0.125),
        ],
    )


@pytest.fixture(scope="module")
def output_options():
    yield OUTPUT_OPTIONS(comment="!", field=10, ndec_block=4, len=20, ndec_spec=6)


@pytest.fixture(scope="module")
def block(times):
    yield BLOCK(
        sname="COMPGRID",
        header=False,
        fname="./output-grid.nc",
        idla=3,
        output=["hsign", "hswell", "dir", "tps", "tm01", "watlev", "qp"],
        times=times,
    )


@pytest.fixture(scope="module")
def table(times):
    yield TABLE(
        sname="outpts",
        format="noheader",
        fname="./output_table.nc",
        output=["hsign", "hswell", "dir", "tps", "tm01", "watlev", "qp"],
        times=times,
    )


@pytest.fixture(scope="module")
def specout(times):
    yield SPECOUT(
        sname="outpts",
        dim=dict(model_type="spec2d"),
        freq=dict(model_type="rel"),
        fname="./specout.nc",
        times=times,
    )


@pytest.fixture(scope="module")
def nestout(times):
    yield NESTOUT(sname="outnest", fname="./nestout.swn", times=times)


@pytest.fixture(scope="module")
def test():
    yield TEST(
        points=dict(
            model_type="xy",
            x=np.linspace(172.5, 174.0, 25),
            y=25 * [-38],
        ),
        fname_s2d="2d_variance_density.test",
    )


def test_base_location():
    loc = BaseLocation(sname="outsites")
    assert loc.render() == "LOCATIONS sname='outsites'"


def test_sname_lt8():
    with pytest.raises(ValidationError):
        BaseLocation(sname="outputlocations")


@pytest.mark.parametrize("sname", SPECIAL_NAMES)
def test_sname_special(sname):
    with pytest.raises(ValidationError):
        BaseLocation(sname=sname)


def test_frame(frame):
    assert frame.render() == (
        "FRAME sname='outgrid' xpfr=173.0 ypfr=-40.0 alpfr=0.0 "
        "xlenfr=2.0 ylenfr=2.0 mxfr=19 myfr=19"
    )


def test_group(group):
    group.render() == "GROUP sname='subgrid' SUBGRID ix1=20 iy1=0 ix2=50 iy2=100"


def test_curve(curve):
    assert "CURVE sname='outcurve'" in curve.render()


def test_curve(curve):
    loc = CURVES(curves=[curve, curve])
    assert loc.render().count("CURVE sname=") == 2


def test_ray(ray):
    assert ray.render().startswith("RAY rname='outray'")


def test_isoline(isoline):
    assert isoline.render() == "ISOLINE sname='outcurve' rname='outray' DEPTH dep=12.0"


def test_points(points):
    assert points.render().startswith("POINTS sname='outpts' &")


def test_points_file():
    loc = POINTS_FILE(sname="outpts", fname="./output_points.nc")
    assert loc.render() == "POINTS sname='outpts' fname='./output_points.nc'"


def test_ngrid(ngrid):
    assert ngrid.render() == (
        "NGRID sname='outnest' xpn=173.0 ypn=-40.0 alpn=0.0 xlenn=2.0 ylenn=2.0 "
        "mxn=19 myn=19"
    )


def test_ngrid_unstructured():
    loc = NGRID_UNSTRUCTURED(sname="outnest", kind="triangle", fname="./ngrid.txt")
    assert loc.render() == (
        "NGRID sname='outnest' UNSTRUCTURED TRIANGLE fname='./ngrid.txt'"
    )


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(output=["xp"], hexp=100),
        dict(output=["hsign", "tm01", "rtmm10"], excv=-9),
        dict(output=["hsign", "tm02", "fspr"], fmin=0.03, fmax=0.5),
        dict(output=["hsign"], fswell=0.08),
        dict(output=["per"], short="Tm-1,0", power=0),
        dict(output=["transp", "force"], coord="frame"),
    ],
)
def test_quantity(kwargs):
    QUANTITY(**kwargs)


def test_quantity_valid_output():
    with pytest.raises(ValidationError):
        QUANTITY(output=["hsign", "notvalid"])


def test_output_options(output_options):
    assert output_options.render() == (
        "OUTPUT OPTIONS comment='!' TABLE field=10 BLOCK ndec=4 len=20 SPEC ndec=6"
    )


def test_block():
    block = BLOCK(sname="outgrid", fname="./depth-frame.nc", output=["depth"])
    block.render() == "BLOCK sname='outgrid' fname='./depth-frame.nc' DEPTH"


def test_blocks():
    block1 = BLOCK(sname="outgrid", fname="./depth.txt", output=["depth"])
    block2 = BLOCK(
        sname="outgrid",
        fname="./output.nc",
        output=["hsign", "hswell", "tm01", "tps", "vel", "wind"],
    )
    blocks = BLOCKS(components=[block1, block2])
    assert len(blocks.components) == 2
    assert isinstance(blocks.components[0], BLOCK)


def test_block_nonstationary(block):
    assert "OUTPUT tbegblk=19900101.000000 deltblk=1.0 HR" in block.render()


def test_table(table):
    assert table.render().startswith(
        "TABLE sname='outpts' NOHEADER fname='./output_table.nc' &"
    )


def test_specout(specout):
    assert specout.render() == (
        "SPECOUT sname='outpts' SPEC2D REL fname='./specout.nc' "
        "OUTPUT tbegspc=19900101.000000 deltspc=1.0 HR"
    )


def test_nestout(nestout):
    assert nestout.render() == (
        "NESTOUT sname='outnest' fname='./nestout.swn' "
        "OUTPUT tbegnst=19900101.000000 deltnst=1.0 HR"
    )


def test_test_xy(test):
    assert test.render().startswith("TEST POINTS XY &")


def test_test_ij():
    test = TEST(
        itest=10,
        points=dict(model_type="ij", i=[0, 0], j=[10, 20]),
        fname_par="integral_parameters.test",
        fname_s1d="1d_variance_density.test",
        fname_s2d="2d_variance_density.test",
    )
    print(test.render())


def test_test_max50():
    with pytest.raises(ValidationError):
        TEST(
            points=dict(
                model_type="xy",
                x=np.linspace(172.5, 174.0, 60),
                y=60 * [-38],
            ),
            fname_s2d="2d_variance_density.test",
        )


def test_output_group_all_set(
    frame,
    group,
    curves,
    ray,
    isoline,
    points,
    ngrid,
    quantities,
    output_options,
    block,
    table,
    specout,
    nestout,
):
    output = OUTPUT(
        frame=frame,
        group=group,
        curve=curves,
        ray=ray,
        isoline=isoline,
        points=points,
        ngrid=ngrid,
        quantity=quantities,
        output_options=output_options,
        block=block,
        table=table,
        specout=specout,
        nestout=nestout,
    )
    print("")
    print(output.render())


def test_output_sname_unique(frame, group):
    group1 = copy.deepcopy(group)
    group1.sname = frame.sname
    with pytest.raises(ValidationError):
        OUTPUT(frame=frame, group=group1)


def test_output_block_frame_or_group(points):
    block = BLOCK(sname="outpts", fname="./depth-frame.nc", output=["depth"])
    with pytest.raises(ValidationError):
        OUTPUT(points=points, block=block)


def test_output_block_frame_or_group(points):
    block = BLOCK(sname="outpts", fname="./depth-frame.nc", output=["depth"])
    with pytest.raises(ValidationError):
        OUTPUT(points=points, block=block)


def test_output_write_locations_exist(table):
    with pytest.raises(ValidationError):
        OUTPUT(table=table)


def test_output_ray_defined_if_isoline(isoline):
    with pytest.raises(ValidationError):
        OUTPUT(isoline=isoline)


def test_output_ray_rname_matches_isoline_rname(isoline, ray):
    isoline.rname = "dummy"
    with pytest.raises(ValidationError):
        OUTPUT(isoline=isoline, ray=ray)


def test_output_ngrid_nestout_defined(ngrid, nestout):
    with pytest.raises(ValidationError):
        OUTPUT(ngrid=ngrid)
        OUTPUT(nestout=nestout)


def test_output_sname_ngrid_nestout_match(ngrid, nestout):
    ngrid.sname = "dummy"
    with pytest.raises(ValidationError):
        OUTPUT(ngrid=ngrid, nestout=nestout)



"""Spectrum subcomponents."""

import logging
from typing import Any, Literal, Optional

from pydantic import Field, model_validator

from rompy_swan.subcomponents.base import BaseSubComponent

logger = logging.getLogger(__name__)


class SPECTRUM(BaseSubComponent):
    """SWAN spectrum specification.

    .. code-block:: text

        ->CIRCLE|SECTOR ([dir1] [dir2]) [mdc] [flow] [fhigh] [msc]

    Notes
    -----

    Directions in the spectra are defined either as a CIRCLE or as a SECTOR. In the
    case of a SECTOR, both `dir1` and `dir2` must be specified. In the case of a
    CIRCLE, neither `dir1` nor `dir2` should be specified.

    At least two of `flow`, `fhigh` and `msc` must be specified in which case the
    third parameter will be calculated by SWAN such that the frequency resolution
    `df/f = 0.1` (10% increments).

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.spectrum import SPECTRUM
        spec = SPECTRUM(mdc=36, flow=0.04, fhigh=1.0)
        print(spec.render())
        spec = SPECTRUM(mdc=36, dir1=0, dir2=180, flow=0.04, msc=31)
        print(spec.render())

    """

    model_type: Literal["spectrum", "SPECTRUM"] = Field(
        default="spectrum", description="Model type discriminator"
    )
    mdc: int = Field(
        description=(
            "Number of meshes in theta-space. In the case of CIRCLE, this is the "
            "number of subdivisions of the 360 degrees of a circle so "
            "`dtheta = [360]/[mdc]` is the spectral directional resolution. In the "
            "case of SECTOR, `dtheta = ([dir2] - [dir1])/[mdc]`. The minimum number "
            "of directional bins is 3 per directional quadrant."
        )
    )
    flow: Optional[float] = Field(
        default=None,
        description=(
            "Lowest discrete frequency that is used in the calculation (in Hz)."
        ),
    )
    fhigh: Optional[float] = Field(
        default=None,
        description=(
            "Highest discrete frequency that is used in the calculation (in Hz)."
        ),
    )
    msc: Optional[int] = Field(
        default=None,
        description=(
            "One less than the number of frequencies. This defines the grid "
            "resolution in frequency-space between the lowest discrete frequency "
            "`flow` and the highest discrete frequency `fhigh`. This resolution is "
            "not constant, since the frequencies are distributed logarithmical: "
            "`fi+1 = yfi` where `y` is a constant. The minimum number of frequencies "
            "is 4"
        ),
        ge=3,
    )
    dir1: Optional[float] = Field(
        default=None,
        description=(
            "The direction of the right-hand boundary of the sector when looking "
            "outward from the sector (required for option SECTOR) in degrees."
        ),
    )
    dir2: Optional[float] = Field(
        default=None,
        description=(
            "The direction of the left-hand boundary of the sector when looking "
            "outward from the sector (required for option SECTOR) in degrees."
        ),
    )

    @model_validator(mode="before")
    @classmethod
    def check_direction_definition(cls, data: Any) -> Any:
        """Check that dir1 and dir2 are specified together."""
        dir1 = data.get("dir1")
        dir2 = data.get("dir2")
        if None in [dir1, dir2] and dir1 != dir2:
            raise ValueError("dir1 and dir2 must be specified together")
        return data

    @model_validator(mode="after")
    def check_frequency_definition(self) -> "SPECTRUM":
        """Check spectral frequencies are prescribed correctly."""
        args = [self.flow, self.fhigh, self.msc]
        if None in args:
            args = [arg for arg in args if arg is not None]
            if len(args) != 2:
                raise ValueError("You must specify at least 2 of [flow, fhigh, msc]")
        if self.flow is not None and self.fhigh is not None and self.flow >= self.fhigh:
            raise ValueError("flow must be less than fhigh")
        return self

    @property
    def dir_sector(self):
        if self.dir1 is None and self.dir2 is None:
            return "CIRCLE"
        else:
            return f"SECTOR {self.dir1} {self.dir2}"

    def cmd(self) -> str:
        repr = f"{self.dir_sector} mdc={self.mdc}"
        if self.flow is not None:
            repr += f" flow={self.flow}"
        if self.fhigh is not None:
            repr += f" fhigh={self.fhigh}"
        if self.msc is not None:
            repr += f" msc={self.msc}"
        return repr


class JONSWAP(BaseSubComponent):
    """Jonswap spectral shape.

    .. code-block:: text

        JONSWAP [gamma]

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.spectrum import JONSWAP
        shape = JONSWAP(gamma=3.3)
        print(shape.render())

    """

    model_type: Literal["jonswap", "JONSWAP"] = Field(
        default="jonswap", description="Model type discriminator"
    )
    gamma: float = Field(
        default=3.3,
        description="Peak enhancement parameter of the JONSWAP spectrum.",
        gt=0.0,
    )

    def cmd(self) -> str:
        return f"{super().cmd()} gamma={self.gamma}"


class TMA(JONSWAP):
    """TMA spectral shape.

    .. code-block:: text

        TMA [gamma] [d]

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.spectrum import TMA
        shape = TMA(gamma=2.0, d=18)
        print(shape.render())

    """

    model_type: Literal["tma", "TMA"] = Field(
        default="tma", description="Model type discriminator"
    )
    d: float = Field(
        description="The reference depth at the wave maker in meters.",
        gt=0.0,
    )

    def cmd(self) -> str:
        return f"{super().cmd()} d={self.d}"


class GAUSS(BaseSubComponent):
    """Gaussian spectral shape.

    .. code-block:: text

        GAUSS [sigfr]

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.spectrum import GAUSS
        shape = GAUSS(sigfr=0.02)
        print(shape.render())

    """

    model_type: Literal["gauss", "GAUSS"] = Field(
        default="gauss", description="Model type discriminator"
    )
    sigfr: float = Field(
        description=(
            "Width of the Gaussian frequency spectrum expressed "
            "as a standard deviation in Hz."
        ),
        gt=0.0,
    )

    def cmd(self) -> str:
        return f"{super().cmd()} sigfr={self.sigfr}"


class PM(BaseSubComponent):
    """Pearson-Moskowitz spectral shape.

    .. code-block:: text

        PM

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.spectrum import PM
        shape = PM()
        print(shape.render())

    """

    model_type: Literal["pm", "PM"] = Field(
        default="pm", description="Model type discriminator"
    )


class BIN(BaseSubComponent):
    """Single frequency bin spectral shape.

    .. code-block:: text

        BIN

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.spectrum import BIN
        shape = BIN()
        print(shape.render())

    """

    model_type: Literal["bin", "BIN"] = Field(
        default="bin", description="Model type discriminator"
    )


class SHAPESPEC(BaseSubComponent):
    """Spectral shape specification.

    .. code-block:: text

        BOUND SHAPESPEC JONSWAP|PM|GAUSS|BIN|TMA PEAK|MEAN DSPR [POWER|DEGREES]

    This command BOUND SHAPESPEC defines the shape of the spectra (both in frequency
    and direction) at the boundary of the computational grid in case of parametric
    spectral input.

    Notes
    -----
    While technically a component `BOUND SHAPESPEC`, this is only intended to be used
    as a subcomponent of the `BOUNDSPEC` component.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.spectrum import SHAPESPEC
        shapespec = SHAPESPEC()
        print(shapespec.render())
        shapespec = SHAPESPEC(
            shape=dict(model_type="tma", gamma=3.1, d=12),
            per_type="mean",
            dspr_type="degrees",
        )
        print(shapespec.render())

    """

    model_type: Literal["shapespec", "SHAPESPEC"] = Field(
        default="shapespec", description="Model type discriminator"
    )
    shape: JONSWAP | PM | GAUSS | BIN | TMA = Field(
        default_factory=JONSWAP,
        description="The spectral shape",
    )
    per_type: Literal["peak", "mean"] = Field(
        default="peak",
        description="The type of characteristic wave period",
    )
    dspr_type: Literal["power", "degrees"] = Field(
        default="power",
        description="The type of directional spreading",
    )

    def cmd(self) -> str:
        repr = (
            f"BOUND SHAPESPEC {self.shape.render()} {self.per_type.upper()} "
            f"DSPR {self.dspr_type.upper()}"
        )
        return repr



"""Input grid for SWAN."""

from abc import ABC
from pathlib import Path
from typing import Literal, Optional

from pydantic import Field, model_validator

from rompy_swan.components.base import BaseComponent
from rompy_swan.subcomponents.readgrid import READINP
from rompy_swan.subcomponents.time import NONSTATIONARY
from rompy_swan.types import GridOptions

# TODO: Components are a bit mixed up here, define them a bit better.

HERE = Path(__file__).parent


class INPGRID(BaseComponent, ABC):
    """SWAN input grid.

    .. code-block:: text

        INPGRID [grid_type] ->REGULAR|CURVILINEAR|UNSTRUCTURED (EXCEPTION [excval]) &
            (NONSTATIONARY [tbeginp] [deltinp] ->SEC|MIN|HR|DAY [tendinp])

    This is the base class for all input grids. It is not meant to be used directly.

    """

    model_type: Literal["inpgrid", "INPGRID"] = Field(
        default="inpgrid",
        description="Model type discriminator",
    )
    grid_type: GridOptions = Field(
        description="Type of the swan input grid, e.g, 'bottom', 'wind', etc",
    )
    excval: Optional[float] = Field(
        default=None,
        description=(
            "Exception value to allow identifying and ignoring certain point inside "
            "the given grid during the computation. If `fac` != 1, `excval` must be "
            "given as `fac` times the exception value"
        ),
    )
    nonstationary: Optional[NONSTATIONARY] = Field(
        default=None,
        description="Nonstationary time specification",
    )
    readinp: READINP = Field(
        description="SWAN input grid file reader specification",
    )

    @model_validator(mode="after")
    def set_nonstat_suffix(self) -> "INPGRID":
        """Set the nonstationary suffix."""
        if self.nonstationary is not None:
            self.nonstationary.suffix = "inp"
        if self.grid_type is not None:
            self.readinp.grid_type = self.grid_type
        return self

    def cmd(self) -> str:
        return f"INPGRID {self.grid_type.upper()}"


class REGULAR(INPGRID):
    """SWAN regular input grid.

    .. code-block:: text

        INPGRID [grid_type] REGULAR [xpinp] [ypinp] [alpinp] [mxinp] [myinp] &
            [dxinp] [dyinp] (EXCEPTION [excval]) &
            (NONSTATIONARY [tbeginp] [deltinp] ->SEC|MIN|HR|DAY [tendinp])
        READGRID [grid_type] [fac] 'fname1' [idla] [nhedf] ([nhedt]) ([nhedvec]) &
            ->FREE|FORMAT|UNFORMATTED ('form'|[idfm])

    This is a group component that includes an `INPGRID` and a `READGRID` component.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.inpgrid import REGULAR
        inpgrid = REGULAR(
            grid_type="bottom",
            excval=-99.0,
            xpinp=172.0,
            ypinp=-41.0,
            alpinp=0.0,
            mxinp=99,
            myinp=99,
            dxinp=0.005,
            dyinp=0.005,
            readinp=dict(fname1="bottom.txt"),
        )
        print(inpgrid.render())
        inpgrid = REGULAR(
            grid_type="wind",
            excval=-99.0,
            xpinp=172.0,
            ypinp=-41.0,
            alpinp=0.0,
            mxinp=99,
            myinp=99,
            dxinp=0.005,
            dyinp=0.005,
            readinp=dict(fname1="wind.txt"),
            nonstationary=dict(
                tbeg="2019-01-01T00:00:00",
                tend="2019-01-07 00:00:00",
                delt=3600,
                dfmt="hr",
            ),
        )
        print(inpgrid.render())

    TODO: Use grid object, requires different grid parameters to be allowed.

    """

    model_type: Literal["regular", "REGULAR"] = Field(
        default="regular",
        description="Model type discriminator",
    )
    xpinp: float = Field(
        description=(
            "Geographic location (x-coordinate) of the origin of the input grid in "
            "problem coordinates (in m) if Cartesian coordinates are used or in "
            "degrees if spherical coordinates are used. In case of spherical "
            "coordinates there is no default"
        ),
    )
    ypinp: float = Field(
        description=(
            "Geographic location (y-coordinate) of the origin of the input grid in "
            "problem coordinates (in m) if Cartesian coordinates are used or in "
            "degrees if spherical coordinates are used. In case of spherical "
            "coordinates there is no default"
        ),
    )
    alpinp: Optional[float] = Field(
        default=0.0,
        description=(
            "Direction of the positive x-axis of the input grid "
            "(in degrees, Cartesian convention)"
        ),
    )
    mxinp: int = Field(
        description=(
            "Number of meshes in x-direction of the input grid (this number is one "
            "less than the number of grid points in this direction)"
        ),
    )
    myinp: int = Field(
        description=(
            "Number of meshes in y-direction of the input grid (this number is one "
            "less than the number of grid points in this direction). In 1D-mode, "
            "`myinp` should be 0"
        ),
    )
    dxinp: float = Field(
        description=(
            "Mesh size in x-direction of the input grid, in m in case of Cartesian "
            "coordinates or in degrees if spherical coordinates are used"
        ),
    )
    dyinp: float = Field(
        description=(
            "Mesh size in y-direction of the input grid, in m in case of Cartesian "
            "coordinates or in degrees if spherical coordinates are used. "
            "In 1D-mode, `dyinp` may have any value"
        ),
    )

    def cmd(self) -> list:
        repr = (
            f"{super().cmd()} REGULAR xpinp={self.xpinp} ypinp={self.ypinp} "
            f"alpinp={self.alpinp} mxinp={self.mxinp} myinp={self.myinp} "
            f"dxinp={self.dxinp} dyinp={self.dyinp}"
        )
        if self.excval is not None:
            repr += f" EXCEPTION excval={self.excval}"
        if self.nonstationary is not None:
            repr += f" {self.nonstationary.render()}"
        repr = [repr] + [self.readinp.render()]
        return repr


class CURVILINEAR(INPGRID):
    """SWAN curvilinear input grid.

    .. code-block:: text

        INPGRID [grid_type] CURVILINEAR [stagrx] [stagry] [mxinp] [myinp] &
            (EXCEPTION [excval]) &
            (NONSTATIONARY [tbeginp] [deltinp] ->SEC|MIN|HR|DAY [tendinp])
        READGRID [grid_type] [fac] 'fname1' [idla] [nhedf] ([nhedt]) ([nhedvec]) &
            ->FREE|FORMAT|UNFORMATTED ('form'|[idfm])

    This is a group component that includes an `INPGRID` and a `READGRID` component.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.inpgrid import CURVILINEAR
        inpgrid = CURVILINEAR(
            grid_type="wind",
            stagrx=0.0,
            stagry=0.0,
            mxinp=199,
            myinp=199,
            excval=-99.0,
            readinp=dict(fname1="wind.txt"),
            nonstationary=dict(
                tbeg="2019-01-01T00:00:00",
                tend="2019-01-07 00:00:00",
                delt=3600,
                dfmt="hr",
            ),
        )
        print(inpgrid.render())

    TODO: Handle (or not) setting default values for mxinp and myinp from cgrid.

    """

    model_type: Literal["curvilinear", "CURVILINEAR"] = Field(
        default="curvilinear", description="Model type discriminator"
    )
    stagrx: float = Field(
        default=0.0,
        description=(
            "Staggered x'-direction with respect to computational grid, e.g., "
            "`stagrx=0.5` means that the input grid points are shifted a half step "
            "in x'-direction; in many flow models x-velocities are defined in points "
            "shifted a half step in x'-direction"
        ),
    )
    stagry: float = Field(
        default=0.0,
        description=(
            "Staggered y'-direction with respect to computational grid, e.g., "
            "`stagry=0.5` means that the input grid points are shifted a half step "
            "in y'-direction; in many flow models y-velocities are defined in points "
            "shifted a half step in y'-direction"
        ),
    )
    mxinp: int = Field(
        description=(
            "Number of meshes in ξ-direction of the input grid (this number is one "
            "less than the number of grid points in this direction)"
        ),
    )
    myinp: int = Field(
        description=(
            "Number of meshes in η-direction of the input grid (this number is one "
            "less than the number of grid points in this direction)"
        ),
    )

    def cmd(self) -> str:
        repr = (
            f"{super().cmd()} CURVILINEAR stagrx={self.stagrx} "
            f"stagry={self.stagry} mxinp={self.mxinp} myinp={self.myinp} "
        )
        if self.excval is not None:
            repr += f" EXCEPTION excval={self.excval}"
        if self.nonstationary is not None:
            repr += f" {self.nonstationary.render()}"
        repr = [repr] + [self.readinp.render()]
        return repr


class UNSTRUCTURED(INPGRID):
    """SWAN unstructured input grid.

    .. code-block:: text

        INPGRID [grid_type] UNSTRUCTURED EXCEPTION [excval]) &
            (NONSTATIONARY [tbeginp] [deltinp] ->SEC|MIN|HR|DAY [tendinp])
        READGRID [grid_type] [fac] 'fname1' [idla] [nhedf] ([nhedt]) ([nhedvec]) &
            ->FREE|FORMAT|UNFORMATTED ('form'|[idfm])

    This is a group component that includes an `INPGRID` and a `READGRID` component.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.inpgrid import UNSTRUCTURED
        inpgrid = UNSTRUCTURED(
            grid_type="bottom",
            excval=-99.0,
            readinp=dict(fname1="bottom.txt"),
            nonstationary=dict(
                tbeg="2019-01-01T00:00:00",
                tend="2019-01-07 00:00:00",
                delt=3600,
                dfmt="hr",
            ),
        )
        print(inpgrid.render())

    """

    model_type: Literal["unstructured", "UNSTRUCTURED"] = Field(
        default="unstructured", description="Model type discriminator"
    )

    def cmd(self) -> str:
        repr = f"{super().cmd()} UNSTRUCTURED"
        if self.excval is not None:
            repr += f" EXCEPTION excval={self.excval}"
        if self.nonstationary is not None:
            repr += f" {self.nonstationary.render()}"
        repr = [repr] + [self.readinp.render()]
        return repr


class WIND(BaseComponent):
    """Constant wind input field.

    .. code-block:: text

        WIND [vel] [dir]

    With this optional command, the user indicates that the wind field is constant.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.inpgrid import WIND
        wind = WIND(vel=10.0, dir=270.0)
        print(wind.render())

    """

    model_type: Literal["wind", "WIND"] = Field(
        default="wind", description="Model type discriminator"
    )
    vel: float = Field(description="Wind velocity at 10 m elevation (m/s)", ge=0.0)
    dir: float = Field(
        description=(
            "Wind direction at 10 m elevation (in degrees, Cartesian or Nautical "
            "convention, see command SET)"
        ),
        ge=-180.0,
        le=360.0,
    )

    def cmd(self):
        return f"WIND vel={self.vel} dir={self.dir}"


class ICE(BaseComponent):
    """Constant wind input field.

    .. code-block:: text

        ICE [aice] [hice]

    With this optional command, the user indicates that one or more ice fields are
    constant.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.inpgrid import ICE
        ice = ICE(aice=0.1, hice=0.1)
        print(ice.render())

    """

    model_type: Literal["ice", "ICE"] = Field(
        default="ice", description="Model type discriminator"
    )
    aice: float = Field(
        description="Areal ice fraction, a number from 0 to 1", ge=0.0, le=1.0
    )
    hice: float = Field(description="Ice thickness (m)", ge=0.0)

    def cmd(self):
        return f"ICE aice={self.aice} hice={self.hice}"



"""SWAN numerics subcomponents."""

from typing import Literal, Optional, Union

from pydantic import Field

from rompy_swan.subcomponents.base import BaseSubComponent
from rompy_swan.subcomponents.time import Delt


class BSBT(BaseSubComponent):
    """BSBT first order propagation scheme.

    .. code-block:: text

        BSTB

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.numerics import BSBT
        scheme = BSBT()
        print(scheme.render())

    """

    model_type: Literal["bsbt", "BSBT"] = Field(
        default="bsbt", description="Model type discriminator"
    )

    def cmd(self) -> str:
        """Command file string for this component."""
        return "BSBT"


class GSE(BaseSubComponent):
    """Garden-sprinkler effect.

    .. code-block:: text

        GSE [waveage] Sec|MIn|HR|DAy

    Garden-sprinkler effect is to be counteracted in the S&L propagation scheme
    (default for nonstationary regular grid computations) or in the propagation
    scheme for unstructured grids by adding a diffusion term to the basic equation.
    This may affect the numerical stability of SWAN.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.numerics import GSE
        scheme = GSE(waveage=dict(delt=86400, dfmt="day"))
        print(scheme.render())

    """

    model_type: Literal["gse", "GSE"] = Field(
        default="gse", description="Model type discriminator"
    )
    waveage: Optional[Delt] = Field(
        default=None,
        description=(
            "The time interval used to determine the diffusion which counteracts the "
            "so-called garden-sprinkler effect. The default value of `waveage` is "
            "zero, i.e. no added diffusion. The value of `waveage` should correspond "
            "to the travel time of the waves over the computational region."
        ),
    )

    def cmd(self) -> str:
        """Command file string for this component."""
        repr = "GSE"
        if self.waveage is not None:
            repr += f" waveage={self.waveage.render()}"
        return repr


class STAT(BaseSubComponent):
    """Computation parameters in stationary computation."""

    model_type: Literal["stat", "STAT"] = Field(
        default="stat", description="Model type discriminator"
    )
    mxitst: Optional[int] = Field(
        default=None,
        description=(
            "The maximum number of iterations for stationary computations. The "
            "computation stops when this number is exceeded (SWAN default:  50)"
        ),
    )
    alfa: Optional[float] = Field(
        default=None,
        description=(
            "Proportionality constant used in the frequency-dependent under-"
            "relaxation technique. Based on experiences, a suggestion for this "
            "parameter is `alfa = 0.01`. In case of diffraction computations, the use "
            "of this parameter is recommended (SWAN default: 0.00)"
        ),
    )

    def cmd(self) -> str:
        """Command file string for this component."""
        repr = "STATIONARY"
        if self.mxitst is not None:
            repr += f" mxitst={self.mxitst}"
        if self.alfa is not None:
            repr += f" alfa={self.alfa}"
        return repr


class NONSTAT(BaseSubComponent):
    """Computation parameters in nonstationary computation."""

    model_type: Literal["nonstat", "NONSTAT"] = Field(
        default="nonstat", description="Model type discriminator"
    )
    mxitns: Optional[int] = Field(
        default=None,
        description=(
            "The maximum number of iterations per time step for nonstationary "
            "computations. The computation moves to the next time step when this "
            "number is exceeded (SWAN default: `mxitns = 1`"
        ),
    )

    def cmd(self) -> str:
        """Command file string for this component."""
        repr = "NONSTATIONARY"
        if self.mxitns is not None:
            repr += f" mxitns={self.mxitns}"
        return repr


class STOPC(BaseSubComponent):
    """Stopping criteria of  Zijlema and Van der Westhuysen (2005).

    .. code-block:: text

        STOPC [dabs] [drel] [curvat] [npnts] ->STAT|NONSTAT [limiter]

    With this option the user can influence the criterion for terminating the iterative
    procedure in the SWAN computations (both stationary and nonstationary). The
    criterion makes use of the second derivative, or curvature, of the iteration curve
    of the significant wave height. As the solution of a simulation approaches full
    convergence, the curvature of the iteration curve will tend to zero. SWAN stops the
    process if the relative change in Hs from one iteration to the next is less than
    `drel` and the curvature of the iteration curve of Hs normalized with Hs is less
    than `curvat` or the absolute change in Hs from one iteration to the next is less
    than `dabs`. Both conditions need to be fulfilled in more than fraction `npnts`
    percent of all wet grid points.

    With respect to the QC modelling, another stopping criteria will be employed.
    Namely, SWAN stops the iteration process if the absolute change in Hs from one
    iterate to another is less than `dabs` * Hinc, where Hinc is the representative
    incident wave height, or the relative change in Hs from one to the next iteration
    is less than `drel`. These criteria must be fulfilled in more than `npnts`
    percent of all active, well-defined points.

    References
    ----------
    - Zijlema, M. and Van der Westhuysen, A. (2005). On convergence behaviour and
      numerical accuracy in stationary SWAN simulations of nearshore wind wave spectra,
      Coastal Engineering, 52(3), p. 337-256.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.numerics import STOPC
        stop = STOPC()
        print(stop.render())
        stop = STOPC(
            dabs=0.005,
            drel=0.01,
            curvat=0.005,
            npnts=99.5,
            mode=dict(model_type="nonstat", mxitns=1),
            limiter=0.1,
        )
        print(stop.render())

    """

    model_type: Literal["stopc", "STOPC"] = Field(
        default="stopc", description="Model type discriminator"
    )
    dabs: Optional[float] = Field(
        default=None,
        description=(
            "Maximum absolute change in Hs from one iteration to the next "
            "(SWAN default: 0.005 [m] or 0.05 [-] in case of QC model)"
        ),
    )
    drel: Optional[float] = Field(
        default=None,
        description=(
            "Maximum relative change in Hs from one iteration to the next "
            "(SWAN default: 0.01 [-])"
        ),
    )
    curvat: Optional[float] = Field(
        default=None,
        description=(
            "Maximum curvature of the iteration curve of Hs normalised with Hs "
            "(SWAN default: 0.005 [-] (not used in the QC model))"
        ),
    )
    npnts: Optional[float] = Field(
        default=None,
        description=(
            "Percentage of points in the computational grid above which the stopping "
            "criteria needs to be satisfied (SWAN default: 99.5 [-])"
        ),
    )
    mode: Optional[Union[STAT, NONSTAT]] = Field(
        default=None,
        description="Termination criteria for stationary or nonstationary runs",
        discriminator="model_type",
    )
    limiter: Optional[float] = Field(
        default=None,
        description=(
            "Determines the maximum change per iteration of the energy density per "
            "spectral-bin given in terms of a fraction of the omni-directional "
            "Phillips level (SWAN default: 0.1)"
        ),
    )

    def cmd(self) -> str:
        """Command file string for this component."""
        repr = "STOPC"
        if self.dabs is not None:
            repr += f" dabs={self.dabs}"
        if self.drel is not None:
            repr += f" drel={self.drel}"
        if self.curvat is not None:
            repr += f" curvat={self.curvat}"
        if self.npnts is not None:
            repr += f" npnts={self.npnts}"
        if self.mode is not None:
            repr += f" {self.mode.render()}"
        if self.limiter is not None:
            repr += f" limiter={self.limiter}"
        return repr


class ACCUR(BaseSubComponent):
    """Stop the iterative procedure.

    .. code-block:: text

        ACCUR [drel] [dhoval] [dtoval] [npnts] ->STAT|NONSTAT [limiter]

    With this option the user can influence the criterion for terminating the iterative
    procedure in the SWAN computations (both stationary and non-stationary modes).
    SWAN stops the iterations if (a), (b) and (c) are all satisfied:

    a) The change in the local significant wave height Hs from one iteration to the
       next is less than (1) fraction `drel` of that height or (2) fraction `dhoval`
       of the average Hs over all grid points.

    b) The change in the local mean wave period Tm01 from one iteration to the next is
       less than (1) fraction `drel` of that period or (2) fraction `dtoval` of the
       average mean wave period over all wet grid points.

    c) Conditions (a) and (b) are fulfilled in more than fraction `npnts%` of all wet
       grid points.

    Note
    ----
    This command has become obsolete in SWAN 41.01. The command STOPC should be used.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.numerics import ACCUR
        accur = ACCUR()
        print(accur.render())
        accur = ACCUR(
            drel=0.01,
            dhoval=0.02,
            dtoval=0.02,
            npnts=98.0,
            mode=dict(model_type="nonstat", mxitns=1),
            limiter=0.1,
        )
        print(accur.render())

    """

    model_type: Literal["accur", "ACCUR"] = Field(
        default="accur", description="Model type discriminator"
    )
    drel: Optional[float] = Field(
        default=None,
        description=(
            "Maximum relative change in Hs or Tm01 from one iteration to the next "
            "(SWAN default: 0.02)"
        ),
    )
    dhoval: Optional[float] = Field(
        default=None,
        description=(
            "Fraction of the average Hs over all wet grid points below which the "
            "the stopping criteria needs to be satisfied (SWAN default: 0.02)"
        ),
    )
    dtoval: Optional[float] = Field(
        default=None,
        description=(
            "Fraction of the average Tm01 over all wet grid points below which the "
            "the stopping criteria needs to be satisfied (SWAN default: 0.02)"
        ),
    )
    npnts: Optional[float] = Field(
        default=None,
        description=(
            "Percentage of points in the computational grid above which the stopping "
            "criteria needs to be satisfied (SWAN default: 98)"
        ),
    )
    mode: Optional[Union[STAT, NONSTAT]] = Field(
        default=None,
        description="Termination criteria for stationary or nonstationary runs",
        discriminator="model_type",
    )
    limiter: Optional[float] = Field(
        default=None,
        description=(
            "Determines the maximum change per iteration of the energy density per "
            "spectral-bin given in terms of a fraction of the omni-directional "
            "Phillips level (SWAN default: 0.1)"
        ),
    )

    def cmd(self) -> str:
        """Command file string for this component."""
        repr = "ACCUR"
        if self.drel is not None:
            repr += f" drel={self.drel}"
        if self.dhoval is not None:
            repr += f" dhoval={self.dhoval}"
        if self.dtoval is not None:
            repr += f" dtoval={self.dtoval}"
        if self.npnts is not None:
            repr += f" npnts={self.npnts}"
        if self.mode is not None:
            repr += f" {self.mode.render()}"
        if self.limiter is not None:
            repr += f" limiter={self.limiter}"
        return repr


class DIRIMPL(BaseSubComponent):
    """Numerical scheme for refraction.

    .. code-block:: text

        DIRIMPL [cdd]

    Examples
    --------
    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.numerics import DIRIMPL
        dirimpl = DIRIMPL()
        print(dirimpl.render())
        dirimpl = DIRIMPL(cdd=0.5)
        print(dirimpl.render())

    """

    model_type: Literal["dirimpl", "DIRIMPL"] = Field(
        default="dirimpl", description="Model type discriminator"
    )
    cdd: Optional[float] = Field(
        default=None,
        description=(
            "A value of `cdd=0` corresponds to a central scheme and has the largest "
            "accuracy (diffusion ≈ 0) but the computation may more easily generate"
            "spurious fluctuations. A value of `cdd=1` corresponds to a first order"
            "upwind scheme and it is more diffusive and therefore preferable if "
            "(strong) gradients in depth or current are present (SWAN default: 0.5)"
        ),
    )

    def cmd(self) -> str:
        """Command file string for this component."""
        repr = "DIRIMPL"
        if self.cdd is not None:
            repr += f" cdd={self.cdd}"
        return repr


class SIGIMPL(BaseSubComponent):
    """Frequency shifting accuracy.

    .. code-block:: text

        SIGIMpl [cfl] [eps2] [outp] [niter]

    Controls the accuracy of computing the frequency shifting and the stopping
    criterion and amount of output for the SIP solver (used in the computations in the
    presence of currents or time varying depth)

    Examples
    --------
    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.numerics import SIGIMPL
        sigimpl = SIGIMPL()
        print(sigimpl.render())
        sigimpl = SIGIMPL(css=0.5, eps2=1e-4, outp=0, niter=20)
        print(sigimpl.render())

    """

    model_type: Literal["sigimpl", "SIGIMPL"] = Field(
        default="sigimpl", description="Model type discriminator"
    )
    css: Optional[float] = Field(
        default=None,
        description=(
            "A value of `css=0` corresponds to a central scheme and has the largest "
            "accuracy (diffusion ≈ 0) but the computation may more easily generate "
            "spurious fluctuations. A value of `css=1` corresponds to a first order "
            "upwind scheme and it is more diffusive and therefore preferable if "
            "(strong) gradients in depth or current are present (SWAN default: 0.5)"
        ),
    )
    eps2: Optional[float] = Field(
        default=None,
        description=(
            "Relative stopping criterion to terminate the linear solver (SIP or SOR). "
            "(SWAN default: 1.e-4 in case of SIP and 1.e-6 in case of SOR)"
        ),
    )
    outp: Optional[Literal[0, 1, 2, 3]] = Field(
        default=None,
        description=(
            "Output for the iterative solver:\n\n* 0 = no output\n* 1 = additional "
            "information about the iteration process is written to the PRINT file "
            "\n* 2 = gives a maximal amount of output concerning the iteration "
            "process\n* 3 = summary of the iteration process\n\n(SWAN default: 0)"
        ),
    )
    niter: Optional[int] = Field(
        default=None,
        description=(
            "Maximum number of iterations for the linear solver (SWAN default: 20 in "
            "case of SIP and 1000 in case of SOR)"
        ),
    )

    def cmd(self) -> str:
        """Command file string for this component."""
        repr = "SIGIMPL"
        if self.css is not None:
            repr += f" css={self.css}"
        if self.eps2 is not None:
            repr += f" eps2={self.eps2}"
        if self.outp is not None:
            repr += f" outp={self.outp}"
        if self.niter is not None:
            repr += f" niter={self.niter}"
        return repr


class CTHETA(BaseSubComponent):
    """Prevents excessive directional turning.

    .. code-block:: text

        CTheta [cfl]

    This option prevents an excessive directional turning at a single grid point or
    vertex due to a very coarse bathymetry or current locally. This option limits the
    directional turning rate cθ based on the CFL restriction. (See Eq. 3.41 of
    Scientific/Technical documentation). See also the final remark in Section 2.6.3.
    Note that if this command is not specified, then the limiter is not activated.

    Examples
    --------
    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.numerics import CTHETA
        ctheta = CTHETA()
        print(ctheta.render())
        ctheta = CTHETA(cfl=0.9)
        print(ctheta.render())

    """

    model_type: Literal["ctheta", "CTHETA"] = Field(
        default="ctheta", description="Model type discriminator"
    )
    cfl: Optional[float] = Field(
        default=None,
        description=(
            "Upper limit for the CFL restriction for ctheta. A suggestion for this "
            "parameter is `cfl = 0.9` (SWAN default: 0.9 when CTHETA is activated)"
        ),
    )

    def cmd(self) -> str:
        """Command file string for this component."""
        repr = "CTHETA"
        if self.cfl is not None:
            repr += f" cfl={self.cfl}"
        return repr


class CSIGMA(BaseSubComponent):
    """Prevents excessive directional turning.

    .. code-block:: text

        CSigma [cfl]

    This option prevents an excessive frequency shifting at a single grid point or
    vertex due to a very coarse bathymetry or current locally. This option limits the
    frequency shifting rate csigma based on the CFL restriction. See also the final
    remark in Section 2.6.3. Note that if this command is not specified, then the
    limiter is not activated.

    Examples
    --------
    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.numerics import CSIGMA
        csigma = CSIGMA()
        print(csigma.render())
        csigma = CSIGMA(cfl=0.9)
        print(csigma.render())

    """

    model_type: Literal["ctheta", "CTHETA"] = Field(
        default="ctheta", description="Model type discriminator"
    )
    cfl: Optional[float] = Field(
        default=None,
        description=(
            "Upper limit for the CFL restriction for csigma. A suggestion for this "
            "parameter is `cfl = 0.9` (SWAN default: 0.9 when CSIGMA is activated)"
        ),
    )

    def cmd(self) -> str:
        """Command file string for this component."""
        repr = "CSIGMA"
        if self.cfl is not None:
            repr += f" cfl={self.cfl}"
        return repr


class SETUP(BaseSubComponent):
    """Stop criteria in the computation of wave setup.

    .. code-block:: text

        SETUP [eps2] [outp] [niter]

    Controls the stopping criterion and amount of output for the SOR solver in the
    computation of the wave-induced set-up.

    Examples
    --------
    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.numerics import SETUP
        setup = SETUP()
        print(setup.render())
        setup = SETUP(eps2=1e-4, outp=0, niter=20)
        print(setup.render())

    """

    model_type: Literal["setup", "SETUP"] = Field(
        default="setup", description="Model type discriminator"
    )
    eps2: Optional[float] = Field(
        default=None,
        description=(
            "Relative stopping criterion to terminate the linear solver (SIP or SOR). "
            "(SWAN default: 1.e-4 in case of SIP and 1.e-6 in case of SOR)"
        ),
    )
    outp: Optional[Literal[0, 1, 2, 3]] = Field(
        default=None,
        description=(
            "Output for the iterative solver:\n\n* 0 = no output\n* 1 = additional "
            "information about the iteration process is written to the PRINT file "
            "\n* 2 = gives a maximal amount of output concerning the iteration process "
            "\n* 3 = summary of the iteration process\n\n(SWAN default: 0)"
        ),
    )
    niter: Optional[int] = Field(
        default=None,
        description=(
            "Maximum number of iterations for the linear solver (SWAN default: 20 in "
            "case of SIP and 1000 in case of SOR)"
        ),
    )

    def cmd(self) -> str:
        """Command file string for this component."""
        repr = "SETUP"
        if self.eps2 is not None:
            repr += f" eps2={self.eps2}"
        if self.outp is not None:
            repr += f" outp={self.outp}"
        if self.niter is not None:
            repr += f" niter={self.niter}"
        return repr
